from gdsfactory.grid import grid
from gdsfactory.cell import cell
from gdsfactory.component import Component, copy
from gdsfactory.components.rectangle import rectangle
from glayout.pdk.mappedpdk import MappedPDK
from typing import Optional, Union
from glayout.primitives.via_gen import via_array, via_stack
from glayout.primitives.guardring import tapring
from pydantic import validate_arguments
from glayout.pdk.util.comp_utils import evaluate_bbox, to_float, to_decimal, prec_array, prec_center, prec_ref_center, movey, align_comp_to_port
from glayout.pdk.util.port_utils import rename_ports_by_orientation, rename_ports_by_list, add_ports_perimeter, print_ports
from glayout.routing.c_route import c_route
from glayout.routing.L_route import L_route
from glayout.pdk.util.snap_to_grid import component_snap_to_grid
from decimal import Decimal
from glayout.routing.straight_route import straight_route
from glayout.spice import Netlist


@validate_arguments
def __gen_fingers_macro(pdk: MappedPDK, rmult: int, fingers: int, length: float, width: float, poly_height: float, sdlayer: str, inter_finger_topmet: str) -> Component:
    """internal use: returns an array of fingers"""
    length = pdk.snap_to_2xgrid(length)
    width = pdk.snap_to_2xgrid(width)
    poly_height = pdk.snap_to_2xgrid(poly_height)
    sizing_ref_viastack = via_stack(pdk, "active_diff", "met1")
    # figure out poly (gate) spacing: s/d metal doesnt overlap transistor, s/d min seperation criteria is met
    sd_viaxdim = rmult*evaluate_bbox(via_stack(pdk, "active_diff", "met1"))[0]
    poly_spacing = 2 * pdk.get_grule("poly", "mcon")["min_separation"] + pdk.get_grule("mcon")["width"]
    poly_spacing = max(sd_viaxdim, poly_spacing)
    met1_minsep = pdk.get_grule("met1")["min_separation"]
    poly_spacing += met1_minsep if length < met1_minsep else 0
    # create a single finger
    finger = Component("finger")
    gate = finger << rectangle(size=(length, poly_height), layer=pdk.get_glayer("poly"), centered=True)
    sd_viaarr = via_array(pdk, "active_diff", "met1", size=(sd_viaxdim, width), minus1=True, lay_bottom=False).copy()
    interfinger_correction = via_array(pdk,"met1",inter_finger_topmet, size=(None, width),lay_every_layer=True, num_vias=(1,None))
    sd_viaarr << interfinger_correction
    sd_viaarr_ref = finger << sd_viaarr
    sd_viaarr_ref.movex((poly_spacing+length) / 2)
    finger.add_ports(gate.get_ports_list(),prefix="gate_")
    finger.add_ports(sd_viaarr_ref.get_ports_list(),prefix="rightsd_")
    # create finger array
    fingerarray = prec_array(finger, columns=fingers, rows=1, spacing=(poly_spacing+length, 1),absolute_spacing=True)
    sd_via_ref_left = fingerarray << sd_viaarr
    sd_via_ref_left.movex(0-(poly_spacing+length)/2)
    fingerarray.add_ports(sd_via_ref_left.get_ports_list(),prefix="leftsd_")
    # center finger array and add ports
    centered_farray = Component()
    fingerarray_ref_center = prec_ref_center(fingerarray)
    centered_farray.add(fingerarray_ref_center)
    centered_farray.add_ports(fingerarray_ref_center.get_ports_list())
    # create diffusion and +doped region
    multiplier = rename_ports_by_orientation(centered_farray)
    diff_extra_enc = 2 * pdk.get_grule("mcon", "active_diff")["min_enclosure"]
    diff_dims =(diff_extra_enc + evaluate_bbox(multiplier)[0], width)
    diff = multiplier << rectangle(size=diff_dims,layer=pdk.get_glayer("active_diff"),centered=True)
    sd_diff_ovhg = pdk.get_grule(sdlayer, "active_diff")["min_enclosure"]
    sdlayer_dims = [dim + 2*sd_diff_ovhg for dim in diff_dims]
    sdlayer_ref = multiplier << rectangle(size=sdlayer_dims, layer=pdk.get_glayer(sdlayer),centered=True)
    multiplier.add_ports(sdlayer_ref.get_ports_list(),prefix="plusdoped_")
    multiplier.add_ports(diff.get_ports_list(),prefix="diff_")
    return component_snap_to_grid(rename_ports_by_orientation(multiplier))

def fet_netlist(
    circuit_name: str,
    model: str,
    width: float,
    length: float,
    fingers: int,
    multipliers: int,
    with_dummy: Union[bool, tuple[bool, bool]]
) -> Netlist:
     # add spice netlist
    num_dummies = 0
    if with_dummy == False or with_dummy == (False, False):
        num_dummies = 0
    elif with_dummy == (True, False) or with_dummy == (False, True):
        num_dummies = 1
    elif with_dummy == True or with_dummy == (True, True):
        num_dummies = 2

    source_netlist=""".subckt {circuit_name} {nodes} l=1 w=1 m=1 dm=1
XMAIN   D G S B {model} l={{l}} w={{w}} m={{m}}"""

    for i in range(num_dummies):
        source_netlist += "\nXDUMMY" + str(i+1) + " B B B B {model} l={{l}} w={{w}} m={{dm}}"

    source_netlist += "\n.ends {circuit_name}"

    return Netlist(
        circuit_name=circuit_name,
        nodes=['D', 'G', 'S', 'B'],
        source_netlist=source_netlist,
        instance_format="X{name} {nodes} {circuit_name} l={length} w={width} m={mult} dm={dummy_mult}",
        parameters={
            'model': model,
            'length': length,
            'width': width,
            'mult': fingers * multipliers,
            'dummy_mult': multipliers
        }
    )

@cell
def multiplier(
    pdk: MappedPDK,
    sdlayer: str,
    width: Optional[float] = 3,
    length: Optional[float] = None,
    fingers: int = 1,
    routing: bool = True,
    inter_finger_topmet: str = "met2",
    dummy: Union[bool, tuple[bool, bool]] = True,
    sd_route_topmet: str = "met2",
    gate_route_topmet: str = "met2",
    rmult: Optional[int]=None,
    sd_rmult: int = 1,
    gate_rmult: int=1,
    interfinger_rmult: int=1,
    sd_route_extension: float = 0,
    gate_route_extension: float = 0,
    dummy_routes: bool=True
) -> Component:
    """Generic poly/sd vias generator
    args:
    pdk = pdk to use
    sdlayer = either p+s/d for pmos or n+s/d for nmos
    width = expands the transistor in the y direction
    length = transitor length (if left None defaults to min length)
    fingers = introduces additional fingers (sharing s/d) of width=width
    routing = true or false, specfies if sd should be connected
    inter_finger_topmet = top metal of the via array laid on the source/drain regions
    ****NOTE: routing metal is layed over the source drain regions regardless of routing option
    dummy = true or false add dummy active/plus doped regions
    sd_rmult = multiplies thickness of sd metal (int only)
    gate_rmult = multiplies gate by adding rows to the gate via array (int only)
    interfinger_rmult = multiplies thickness of source/drain routes between the gates (int only)
    sd_route_extension = float, how far extra to extend the source/drain connections (default=0)
    gate_route_extension = float, how far extra to extend the gate connection (default=0)
    dummy_routes: bool default=True, if true add add vias and short dummy poly,source,drain

    ports (one port for each edge),
    ****NOTE: source is below drain:
    gate_... all edges (top met route of gate connection)
    source_...all edges (top met route of source connections)
    drain_...all edges (top met route of drain connections)
    plusdoped_...all edges (area of p+s/d or n+s/d layer)
    diff_...all edges (diffusion region)
    rowx_coly_...all ports associated with finger array include gate_... and array_ (array includes all ports of the viastacks in the array)
    leftsd_...all ports associated with the left most via array
    dummy_L,R_N,E,S,W ports if dummy_routes=True
    """
    # error checking
    if "+s/d" not in sdlayer:
        raise ValueError("specify + doped region for multiplier")
    if not "met" in sd_route_topmet or not "met" in gate_route_topmet:
        raise ValueError("topmet specified must be metal layer")
    if rmult:
        if rmult<1:
            raise ValueError("rmult must be positive int")
        sd_rmult = rmult
        gate_rmult = 1
        interfinger_rmult = ((rmult-1) or 1)
    if sd_rmult<1 or interfinger_rmult<1 or gate_rmult<1:
        raise ValueError("routing multipliers must be positive int")
    if fingers < 1:
        raise ValueError("number of fingers must be positive int")
    # argument parsing and rule setup
    min_length = pdk.get_grule("poly")["min_width"]
    length = min_length if (length or min_length) <= min_length else length
    length = pdk.snap_to_2xgrid(length)
    min_width = max(min_length, pdk.get_grule("active_diff")["min_width"])
    width = min_width if (width or min_width) <= min_width else width
    width = pdk.snap_to_2xgrid(width)
    poly_height = width + 2 * pdk.get_grule("poly", "active_diff")["overhang"]
    # call finger array
    multiplier = __gen_fingers_macro(pdk, interfinger_rmult, fingers, length, width, poly_height, sdlayer, inter_finger_topmet)
    # route all drains/ gates/ sources
    if routing:
        # place vias, then straight route from top port to via-botmet_N
        sd_N_port = multiplier.ports["leftsd_top_met_N"]
        sdvia = via_stack(pdk, "met1", sd_route_topmet)
        sdmet_hieght = sd_rmult*evaluate_bbox(sdvia)[1]
        sdroute_minsep = pdk.get_grule(sd_route_topmet)["min_separation"]
        sdvia_ports = list()
        for finger in range(fingers+1):
            diff_top_port = movey(sd_N_port,destination=width/2)
            # place sdvia such that metal does not overlap diffusion
            big_extension = sdroute_minsep + sdmet_hieght/2 + sdmet_hieght
            sdvia_extension = big_extension if finger % 2 else sdmet_hieght/2
            sdvia_ref = align_comp_to_port(sdvia,diff_top_port,alignment=('c','t'))
            multiplier.add(sdvia_ref.movey(sdvia_extension + pdk.snap_to_2xgrid(sd_route_extension)))
            multiplier << straight_route(pdk, diff_top_port, sdvia_ref.ports["bottom_met_N"])
            sdvia_ports += [sdvia_ref.ports["top_met_W"], sdvia_ref.ports["top_met_E"]]
            # get the next port (break before this if last iteration because port D.N.E. and num gates=fingers)
            if finger==fingers:
                break
            sd_N_port = multiplier.ports[f"row0_col{finger}_rightsd_top_met_N"]
            # route gates
            gate_S_port = multiplier.ports[f"row0_col{finger}_gate_S"]
            metal_seperation = pdk.util_max_metal_seperation()
            psuedo_Ngateroute = movey(gate_S_port.copy(),0-metal_seperation-gate_route_extension)
            psuedo_Ngateroute.y = pdk.snap_to_2xgrid(psuedo_Ngateroute.y)
            multiplier << straight_route(pdk,gate_S_port,psuedo_Ngateroute)
        # place route met: gate
        gate_width = gate_S_port.center[0] - multiplier.ports["row0_col0_gate_S"].center[0] + gate_S_port.width
        gate = rename_ports_by_list(via_array(pdk,"poly",gate_route_topmet, size=(gate_width,None),num_vias=(None,gate_rmult), no_exception=True, fullbottom=True),[("top_met_","gate_")])
        gate_ref = align_comp_to_port(gate.copy(), psuedo_Ngateroute, alignment=(None,'b'),layer=pdk.get_glayer("poly"))
        multiplier.add(gate_ref)
        # place route met: source, drain
        sd_width = sdvia_ports[-1].center[0] - sdvia_ports[0].center[0]
        sd_route = rectangle(size=(sd_width,sdmet_hieght),layer=pdk.get_glayer(sd_route_topmet),centered=True)
        source = align_comp_to_port(sd_route.copy(), sdvia_ports[0], alignment=(None,'c'))
        drain = align_comp_to_port(sd_route.copy(), sdvia_ports[2], alignment=(None,'c'))
        multiplier.add(source)
        multiplier.add(drain)
        # add ports
        multiplier.add_ports(drain.get_ports_list(), prefix="drain_")
        multiplier.add_ports(source.get_ports_list(), prefix="source_")
        multiplier.add_ports(gate_ref.get_ports_list(prefix="gate_"))
    # create dummy regions
    if isinstance(dummy, bool):
        dummyl = dummyr = dummy
    else:
        dummyl, dummyr = dummy
    if dummyl or dummyr:
        dummy = __gen_fingers_macro(pdk,rmult=interfinger_rmult,fingers=1,length=length,width=width,poly_height=poly_height,sdlayer=sdlayer,inter_finger_topmet="met1")
        dummyvia = dummy << via_stack(pdk,"poly","met1",fullbottom=True)
        align_comp_to_port(dummyvia,dummy.ports["row0_col0_gate_S"],layer=pdk.get_glayer("poly"))
        dummy << L_route(pdk,dummyvia.ports["top_met_W"],dummy.ports["leftsd_top_met_S"])
        dummy << L_route(pdk,dummyvia.ports["top_met_E"],dummy.ports["row0_col0_rightsd_top_met_S"])
        dummy.add_ports(dummyvia.get_ports_list(),prefix="gsdcon_")
        dummy_space = pdk.get_grule(sdlayer)["min_separation"] + dummy.xmax
        sides = list()
        if dummyl:
            sides.append((-1,"dummy_L_"))
        if dummyr:
            sides.append((1,"dummy_R_"))
        for side, name in sides:
            dummy_ref = multiplier << dummy
            dummy_ref.movex(side * (dummy_space + multiplier.xmax))
            multiplier.add_ports(dummy_ref.get_ports_list(),prefix=name)
    # ensure correct port names and return
    return component_snap_to_grid(rename_ports_by_orientation(multiplier))


@validate_arguments
def __mult_array_macro(
    pdk: MappedPDK,
    sdlayer: str,
    width: Optional[float] = 3,
    fingers: Optional[int] = 1,
    multipliers: Optional[int] = 1,
    routing: Optional[bool] = True,
    dummy: Optional[Union[bool, tuple[bool, bool]]] = True,
    length: Optional[float] = None,
    sd_route_topmet: Optional[str] = "met2",
    gate_route_topmet: Optional[str] = "met2",
    sd_route_left: Optional[bool] = True,
    sd_rmult: int = 1,
    gate_rmult: int=1,
    interfinger_rmult: int=1,
    dummy_routes: bool=True
) -> Component:
    """create a multiplier array with multiplier_0 at the bottom
    The array is correctly centered
    """
    # create multiplier array
    pdk.activate()
    # TODO: error checking
    multiplier_arr = Component("temp multiplier array")
    multiplier_comp = multiplier(
        pdk,
        sdlayer,
        width=width,
        fingers=fingers,
        dummy=dummy,
        routing=routing,
        length=length,
        sd_route_topmet=sd_route_topmet,
        gate_route_topmet=gate_route_topmet,
        sd_rmult=sd_rmult,
        gate_rmult=gate_rmult,
        interfinger_rmult=interfinger_rmult,
        dummy_routes=dummy_routes
    )
    _max_metal_seperation_ps = max([pdk.get_grule("met"+str(i))["min_separation"] for i in range(1,5)])
    multiplier_separation = (
        to_decimal(_max_metal_seperation_ps)
        + evaluate_bbox(multiplier_comp, True)[1]
    )
    for rownum in range(multipliers):
        row_displacment = rownum * multiplier_separation - (multiplier_separation/2 * (multipliers-1))
        row_ref = multiplier_arr << multiplier_comp
        row_ref.movey(to_float(row_displacment))
        multiplier_arr.add_ports(
            row_ref.get_ports_list(), prefix="multiplier_" + str(rownum) + "_"
        )
    # TODO: fix extension (both extension are broken. IDK src extension and drain extension IDK metal layer)
    src_extension = to_decimal(0.6)
    drain_extension = src_extension + 3*to_decimal(pdk.get_grule("met4")["min_separation"])
    sd_side = "W" if sd_route_left else "E"
    gate_side = "E" if sd_route_left else "W"
    if routing and multipliers > 1:
        for rownum in range(multipliers-1):
            thismult = "multiplier_" + str(rownum) + "_"
            nextmult = "multiplier_" + str(rownum+1) + "_"
            # route sources left
            srcpfx = thismult + "source_"
            this_src = multiplier_arr.ports[srcpfx+sd_side]
            next_src = multiplier_arr.ports[nextmult + "source_"+sd_side]
            src_ref = multiplier_arr << c_route(pdk, this_src, next_src, viaoffset=(True,False), extension=to_float(src_extension))
            multiplier_arr.add_ports(src_ref.get_ports_list(), prefix=srcpfx)
            # route drains left
            drainpfx = thismult + "drain_"
            this_drain = multiplier_arr.ports[drainpfx+sd_side]
            next_drain = multiplier_arr.ports[nextmult + "drain_"+sd_side]
            drain_ref = multiplier_arr << c_route(pdk, this_drain, next_drain, viaoffset=(True,False), extension=to_float(drain_extension))
            multiplier_arr.add_ports(drain_ref.get_ports_list(), prefix=drainpfx)
            # route gates right
            gatepfx = thismult + "gate_"
            this_gate = multiplier_arr.ports[gatepfx+gate_side]
            next_gate = multiplier_arr.ports[nextmult + "gate_"+gate_side]
            gate_ref = multiplier_arr << c_route(pdk, this_gate, next_gate, viaoffset=(True,False), extension=to_float(src_extension))
            multiplier_arr.add_ports(gate_ref.get_ports_list(), prefix=gatepfx)
    multiplier_arr = component_snap_to_grid(rename_ports_by_orientation(multiplier_arr))
    # recenter
    final_arr = Component()
    marrref = final_arr << multiplier_arr
    correctionxy = prec_center(marrref)
    marrref.movex(correctionxy[0]).movey(correctionxy[1])
    final_arr.add_ports(marrref.get_ports_list())
    return component_snap_to_grid(rename_ports_by_orientation(final_arr))


@cell
def nmos(
    pdk,
    width: float = 3,
    fingers: Optional[int] = 1,
    multipliers: Optional[int] = 1,
    with_tie: bool = True,
    with_dummy: Union[bool, tuple[bool, bool]] = True,
    with_dnwell: bool = True,
    with_substrate_tap: bool = True,
    length: Optional[float] = None,
    sd_route_topmet: str = "met2",
    gate_route_topmet: str = "met2",
    sd_route_left: bool = True,
    rmult: Optional[int] = None,
    sd_rmult: int=1,
    gate_rmult: int=1,
    interfinger_rmult: int=1,
    tie_layers: tuple[str,str] = ("met2","met1"),
    substrate_tap_layers: tuple[str,str] = ("met2","met1"),
    dummy_routes: bool=True
) -> Component:
    """Generic NMOS generator
    pdk: mapped pdk to use
    width: expands the NMOS in the y direction
    fingers: introduces additional fingers (sharing source/drain) of width=width
    multipliers: number of multipliers (a multiplier is a row of fingers)
    with_tie: true or false, specfies if a bulk tie is required
    with_dummy: tuple(bool,bool) or bool specifying both sides dummy or neither side dummy
    ****using the tuple option, you can specify a single side dummy such as true,false
    with_dnwell: bool use dnwell (multi well)
    with_substrate_tap: add substrate tap on the very outside perimeter of nmos
    length: if None or below min_length will default to min_length
    sd_route_topmet: specify top metal glayer for the source/drain route
    gate_route_topmet: specify top metal glayer for the gate route
    sd_route_left: specify if the source/drain inter-multiplier routes should be on the left side or right side (if false)
    rmult: if not None overrides all other multiplier options to provide a simple routing multiplier (int only)
    sd_rmult: mulitplies the thickness of the source drain route (int only)
    gate_rmult: add additional via rows to the gate route via array (int only)
    interfinger_rmult: multiplies the thickness of the metal routes between the fingers (int only)
    tie_layers: tuple[str,str] specifying (horizontal glayer, vertical glayer) or well tie ring. default=("met2","met1")
    substrate_tap_layers: tuple[str,str] specifying (horizontal glayer, vertical glayer) or substrate tap ring. default=("met2","met1")
    dummy_routes: bool default=True, if true add add vias and short dummy poly,source,drain
    """
    # TODO: glayer checks
    pdk.activate()
    nfet = Component()
    if rmult:
        if rmult<1:
            raise ValueError("rmult must be positive int")
        sd_rmult = rmult
        gate_rmult = 1
        interfinger_rmult = ((rmult-1) or 1)
    # create and add multipliers to nfet
    multiplier_arr = __mult_array_macro(
        pdk,
        "n+s/d",
        width,
        fingers,
        multipliers,
        dummy=with_dummy,
        length=length,
        sd_route_topmet=sd_route_topmet,
        gate_route_topmet=gate_route_topmet,
        sd_route_left=sd_route_left,
        sd_rmult=sd_rmult,
        gate_rmult=gate_rmult,
        interfinger_rmult=interfinger_rmult,
        dummy_routes=dummy_routes
    )
    multiplier_arr_ref = multiplier_arr.ref()
    nfet.add(multiplier_arr_ref)
    nfet.add_ports(multiplier_arr_ref.get_ports_list())
    # add tie if tie
    if with_tie:
        tap_separation = max(
            pdk.util_max_metal_seperation(),
            pdk.get_grule("active_diff", "active_tap")["min_separation"],
        )
        tap_separation += pdk.get_grule("p+s/d", "active_tap")["min_enclosure"]
        tap_encloses = (
            2 * (tap_separation + nfet.xmax),
            2 * (tap_separation + nfet.ymax),
        )
        tiering_ref = nfet << tapring(
            pdk,
            enclosed_rectangle=tap_encloses,
            sdlayer="p+s/d",
            horizontal_glayer=tie_layers[0],
            vertical_glayer=tie_layers[1],
        )
        nfet.add_ports(tiering_ref.get_ports_list(), prefix="tie_")
        for row in range(multipliers):
            for dummyside,tieside in [("L","W"),("R","E")]:
                try:
                    nfet<<straight_route(pdk,nfet.ports[f"multiplier_{row}_dummy_{dummyside}_gsdcon_top_met_W"],nfet.ports[f"tie_{tieside}_top_met_{tieside}"],glayer2="met1")
                except KeyError:
                    pass
    # add pwell
    nfet.add_padding(
        layers=(pdk.get_glayer("pwell"),),
        default=pdk.get_grule("pwell", "active_tap")["min_enclosure"],
    )
    nfet = add_ports_perimeter(nfet,layer=pdk.get_glayer("pwell"),prefix="well_")
    # add dnwell if dnwell
    if with_dnwell:
        nfet.add_padding(
            layers=(pdk.get_glayer("dnwell"),),
            default=pdk.get_grule("pwell", "dnwell")["min_enclosure"],
        )
    # add substrate tap if with_substrate_tap
    if with_substrate_tap:
        substrate_tap_separation = pdk.get_grule("dnwell", "active_tap")[
            "min_separation"
        ]
        substrate_tap_encloses = (
            2 * (substrate_tap_separation + nfet.xmax),
            2 * (substrate_tap_separation + nfet.ymax),
        )
        ringtoadd = tapring(
            pdk,
            enclosed_rectangle=substrate_tap_encloses,
            sdlayer="p+s/d",
            horizontal_glayer=substrate_tap_layers[0],
            vertical_glayer=substrate_tap_layers[1],
        )
        tapring_ref = nfet << ringtoadd
        nfet.add_ports(tapring_ref.get_ports_list(),prefix="guardring_")

    component = rename_ports_by_orientation(nfet).flatten()

    component.info['netlist'] = fet_netlist(
        circuit_name="NMOS",
        model=pdk.models['nfet'],
        width=width,
        length=length,
        fingers=fingers,
        multipliers=multipliers,
        with_dummy=with_dummy
    )

    return component


@cell
def pmos(
    pdk,
    width: float = 3,
    fingers: Optional[int] = 1,
    multipliers: Optional[int] = 1,
    with_tie: Optional[bool] = True,
    dnwell: Optional[bool] = False,
    with_dummy: Optional[Union[bool, tuple[bool, bool]]] = True,
    with_substrate_tap: Optional[bool] = True,
    length: Optional[float] = None,
    sd_route_topmet: Optional[str] = "met2",
    gate_route_topmet: Optional[str] = "met2",
    sd_route_left: Optional[bool] = True,
    rmult: Optional[int] = None,
    sd_rmult: int=1,
    gate_rmult: int=1,
    interfinger_rmult: int=1,
    tie_layers: tuple[str,str] = ("met2","met1"),
    substrate_tap_layers: tuple[str,str] = ("met2","met1"),
    dummy_routes: bool=True
) -> Component:
    """Generic PMOS generator
    pdk: mapped pdk to use
    width: expands the PMOS in the y direction
    fingers: introduces additional fingers (sharing source/drain) of width=width
    multipliers: number of multipliers (a multiplier is a row of fingers)
    with_tie: true or false, specfies if a bulk tie is required
    dnwell: bool use dnwell if True, or use nwell if False
    with_dummy: tuple(bool,bool) or bool specifying both sides dummy or neither side dummy
    ****using the tuple option, you can specify a single side dummy such as true,false
    with_substrate_tap: add substrate tap on the very outside perimeter of pmos
    length: if None or below min_length will default to min_length
    sd_route_topmet: specify top metal glayer for the source/drain route
    gate_route_topmet: specify top metal glayer for the gate route
    sd_route_left: specify if the source/drain inter-multiplier routes should be on the left side or right side (if false)
    rmult: if not None overrides all other multiplier options to provide a simple routing multiplier (int only)
    sd_rmult: mulitplies the thickness of the source drain route (int only)
    gate_rmult: add additional via rows to the gate route via array (int only)
    interfinger_rmult: multiplies the thickness of the metal routes between the fingers (int only)
    tie_layers: tuple[str,str] specifying (horizontal glayer, vertical glayer) or well tie ring. default=("met2","met1")
    substrate_tap_layers: tuple[str,str] specifying (horizontal glayer, vertical glayer) or substrate tap ring. default=("met2","met1")
    dummy_routes: bool default=True, if true add add vias and short dummy poly,source,drain
    """
    # TODO: glayer checks
    pdk.activate()
    pfet = Component()
    if rmult:
        if rmult<1:
            raise ValueError("rmult must be positive int")
        sd_rmult = rmult
        gate_rmult = 1
        interfinger_rmult = ((rmult-1) or 1)
    # create and add multipliers to nfet
    multiplier_arr = __mult_array_macro(
        pdk,
        "p+s/d",
        width,
        fingers,
        multipliers,
        dummy=with_dummy,
        length=length,
        sd_route_topmet=sd_route_topmet,
        gate_route_topmet=gate_route_topmet,
        sd_route_left=sd_route_left,
        gate_rmult=gate_rmult,
        interfinger_rmult=interfinger_rmult,
        sd_rmult=sd_rmult,
        dummy_routes=dummy_routes
    )
    multiplier_arr_ref = multiplier_arr.ref()
    pfet.add(multiplier_arr_ref)
    pfet.add_ports(multiplier_arr_ref.get_ports_list())
    # add tie if tie
    if with_tie:
        tap_separation = max(
            pdk.get_grule("met2")["min_separation"],
            pdk.get_grule("met1")["min_separation"],
            pdk.get_grule("active_diff", "active_tap")["min_separation"],
        )
        tap_separation += pdk.get_grule("n+s/d", "active_tap")["min_enclosure"]
        tap_encloses = (
            2 * (tap_separation + pfet.xmax),
            2 * (tap_separation + pfet.ymax),
        )
        tapring_ref = pfet << tapring(
            pdk,
            enclosed_rectangle=tap_encloses,
            sdlayer="n+s/d",
            horizontal_glayer=tie_layers[0],
            vertical_glayer=tie_layers[1],
        )
        pfet.add_ports(tapring_ref.get_ports_list(),prefix="tie_")
        for row in range(multipliers):
            for dummyside,tieside in [("L","W"),("R","E")]:
                try:
                    pfet<<straight_route(pdk,pfet.ports[f"multiplier_{row}_dummy_{dummyside}_gsdcon_top_met_W"],pfet.ports[f"tie_{tieside}_top_met_{tieside}"],glayer2="met1")
                except KeyError:
                    pass
    # add nwell
    nwell_glayer = "dnwell" if dnwell else "nwell"
    pfet.add_padding(
        layers=(pdk.get_glayer(nwell_glayer),),
        default=pdk.get_grule("active_tap", nwell_glayer)["min_enclosure"],
    )
    pfet = add_ports_perimeter(pfet,layer=pdk.get_glayer(nwell_glayer),prefix="well_")
    # add substrate tap if with_substrate_tap
    if with_substrate_tap:
        substrate_tap_separation = pdk.get_grule("dnwell", "active_tap")[
            "min_separation"
        ]
        substrate_tap_encloses = (
            2 * (substrate_tap_separation + pfet.xmax),
            2 * (substrate_tap_separation + pfet.ymax),
        )
        pfet << tapring(
            pdk,
            enclosed_rectangle=substrate_tap_encloses,
            sdlayer="p+s/d",
            horizontal_glayer=substrate_tap_layers[0],
            vertical_glayer=substrate_tap_layers[1],
        )
    component =  rename_ports_by_orientation(pfet).flatten()

    component.info['netlist'] = fet_netlist(
        circuit_name="PMOS",
        model=pdk.models['pfet'],
        width=width,
        length=length,
        fingers=fingers,
        multipliers=multipliers,
        with_dummy=with_dummy
    )

    return component





"""read rule deck we have saved in google sheets into a better format for writing to python:
this is the google sheets format
https://docs.google.com/spreadsheets/d/172P-CW_dGQU5icyUQAasLVq7Jlz51R5lB1rJ4JyvfIc/edit?usp=sharing

directions
1) go to the google sheets and download as .csv
2) run this program with the csv input
"""

import csv
from pathlib import Path
from pydantic import validate_arguments


def split_rule(rule: str) -> tuple:
    """Accepts a rule in the expected format and splits into rule name and float value"""
    if (rule != "") and (not "," in rule):
        raise ValueError("rule may be formatted wrong " + rule)
    rule = rule.replace(" ", "").split(",", maxsplit=1)[-1]
    rtr = rule.split("=")
    if len(rtr) != 2:
        rtr.append("*****FIXTHIS!!!MANUALLY!*****")
    elif "," in rtr[1]:
        strlist = rtr[1].replace("(", "").replace(")", "").split(",")
        rtr[1] = tuple([int(layint) for layint in strlist])
    else:
        rtr[1] = float(rtr[1])
    return tuple(rtr)


def __str_rules(groupdata: tuple, group: list, glayers: list) -> str:
    """appends the rules in groupdata to output"""
    group_rules = str()
    for edgenum, edge in enumerate(groupdata):
        group_rules += 'grulesobj["' + group[1] + '"]'
        group_rules += '["' + glayers[edgenum] + '"] = ' + str(edge)
        group_rules += "\n"
    return group_rules


def create_ruledeck_python_dictionary_definition(csvtoread: Path):
    if not csvtoread.is_file():
        raise RuntimeError("csv to read must be a file")
    output = str()
    glayers = list()
    groupdata = []  # list of dictionary
    group = [0, "none"]
    # int,string -> current group
    with open(csvtoread, newline="") as csvfile:
        myreader = csv.reader(csvfile, delimiter=",")
        for rownum, row in enumerate(myreader):
            # deal with header and ignore label rows
            if rownum == 0:
                glayers = row
                glayers.pop(0)
            elif rownum < 3:
                continue
            # processing logic
            # the google sheets csv format is in rows of 3
            # we use group and groupdata to track all rules in a row then update output
            else:
                if group[0] == 0:  # first in group
                    groupdata.clear()
                    group[1] = row[0]
                    for colnum, col in enumerate(row):
                        if colnum == 0:
                            continue
                        groupdata.append(dict())
                for colnum, col in enumerate(row):
                    if colnum == 0:
                        continue
                    if col == "":
                        continue
                    key_val_pair = split_rule(col)
                    groupdata[colnum - 1][key_val_pair[0]] = key_val_pair[1]
                # finished with the group
                if group[0] == 2:  # last in group
                    output += __str_rules(groupdata, group, glayers)
                # update group index
                group[0] = (group[0] + 1) % 3
        # incase missed last group print one more time
        last_grp_rules = __str_rules(groupdata, group, glayers)
        output += "\n" if last_grp_rules in output else last_grp_rules
    return output


@validate_arguments
def visualize_ruleset(ruleset: dict):
    """use networkx to print a visual of the ruleset graph
    nodes are glayers (strings)
    edges are rules (tuple[name:value])
    """
    import pygraphviz as pgv
    ruleGraph = pgv.AGraph(strict=False, directed=False, multigraph=True)
    ruleGraph.add_nodes_from(ruleset.keys())
    for glayer1 in ruleset.keys():
        for glayer2 in ruleset[glayer1].keys():
            for rule, value in ruleset[glayer1][glayer2].items():
                color="black"
                if "min_separ" in rule:
                    color="red"
                elif "min_enclos" in rule:
                    color = "blue"
                elif "width" in rule:
                    color = "orange"
                #ruleGraph.add_edge(glayer1,glayer2,label=value,color=color,key=rule)
                ruleGraph.add_edge(glayer1,glayer2,color=color,key=rule)
    ruleGraph.draw("test.png",format="png",prog="dot")


if __name__ == "__main__":
    from argparse import ArgumentParser

    parser = ArgumentParser(
        prog="print rules", description="read rule deck we have saved in google sheets"
    )
    parser.add_argument("-f", "--file", help="path of csv file to read")
    parser.add_argument(
        "-c",
        "--code",
        action="store_true",
        help="true/false write python file to current dir",
    )
    args = parser.parse_args()
    csvtoread = Path(args.file).resolve()
    output = create_ruledeck_python_dictionary_definition(csvtoread)
    print(output)
    if args.code:
        append_front = """from pdk.mappedpdk import MappedPDK\n
grulesobj = dict()
for glayer in MappedPDK.valid_glayers:
    grulesobj[glayer] = dict((x, None) for x in MappedPDK.valid_glayers)\n
"""
        output = append_front + output
        with open("grules.py", "w") as outputpy:
            outputpy.write(output)



from gdsfactory.cell import cell
from gdsfactory.component import Component, copy
from gdsfactory.components.rectangle import rectangle
from glayout.primitives.fet import nmos, pmos
from glayout.pdk.mappedpdk import MappedPDK
from typing import Optional, Union
from gdsfactory.routing.route_quad import route_quad
from gdsfactory.routing.route_sharp import route_sharp
from glayout.routing.c_route import c_route
from glayout.routing.straight_route import straight_route
from glayout.pdk.util.comp_utils import movex, movey, evaluate_bbox, align_comp_to_port
from glayout.pdk.util.port_utils import rename_ports_by_orientation, rename_ports_by_list, add_ports_perimeter, print_ports, get_orientation, set_port_orientation
from glayout.primitives.via_gen import via_stack
from glayout.primitives.guardring import tapring
from glayout.pdk.util.snap_to_grid import component_snap_to_grid
from glayout.spice import Netlist

def diff_pair_netlist(fetL: Component, fetR: Component) -> Netlist:
	diff_pair_netlist = Netlist(circuit_name='DIFF_PAIR', nodes=['VP', 'VN', 'VDD1', 'VDD2', 'VTAIL', 'B'])
	diff_pair_netlist.connect_netlist(
		fetL.info['netlist'],
		[('D', 'VDD1'), ('G', 'VP'), ('S', 'VTAIL'), ('B', 'B')]
	)
	diff_pair_netlist.connect_netlist(
		fetR.info['netlist'],
		[('D', 'VDD2'), ('G', 'VN'), ('S', 'VTAIL'), ('B', 'B')]
	)

	return diff_pair_netlist

@cell
def diff_pair(
	pdk: MappedPDK,
	width: float = 3,
	fingers: int = 4,
	length: Optional[float] = None,
	n_or_p_fet: bool = True,
	plus_minus_seperation: float = 0,
	rmult: int = 1,
	dummy: Union[bool, tuple[bool, bool]] = True,
	substrate_tap: bool=True
) -> Component:
	"""create a diffpair with 2 transistors placed in two rows with common centroid place. Sources are shorted
	width = width of the transistors
	fingers = number of fingers in the transistors (must be 2 or more)
	length = length of the transistors, None or 0 means use min length
	short_source = if true connects source of both transistors
	n_or_p_fet = if true the diffpair is made of nfets else it is made of pfets
	substrate_tap: if true place a tapring around the diffpair (connects on met1)
	"""
	# TODO: error checking
	pdk.activate()
	diffpair = Component()
	# create transistors
	well = None
	if isinstance(dummy, bool):
		dummy = (dummy, dummy)
	if n_or_p_fet:
		fetL = nmos(pdk, width=width, fingers=fingers,length=length,multipliers=1,with_tie=False,with_dummy=(dummy[0], False),with_dnwell=False,with_substrate_tap=False,rmult=rmult)
		fetR = nmos(pdk, width=width, fingers=fingers,length=length,multipliers=1,with_tie=False,with_dummy=(False,dummy[1]),with_dnwell=False,with_substrate_tap=False,rmult=rmult)
		min_spacing_x = pdk.get_grule("n+s/d")["min_separation"] - 2*(fetL.xmax - fetL.ports["multiplier_0_plusdoped_E"].center[0])
		well = "pwell"
	else:
		fetL = pmos(pdk, width=width, fingers=fingers,length=length,multipliers=1,with_tie=False,with_dummy=(dummy[0], False),dnwell=False,with_substrate_tap=False,rmult=rmult)
		fetR = pmos(pdk, width=width, fingers=fingers,length=length,multipliers=1,with_tie=False,with_dummy=(False,dummy[1]),dnwell=False,with_substrate_tap=False,rmult=rmult)
		min_spacing_x = pdk.get_grule("p+s/d")["min_separation"] - 2*(fetL.xmax - fetL.ports["multiplier_0_plusdoped_E"].center[0])
		well = "nwell"
	# place transistors
	viam2m3 = via_stack(pdk,"met2","met3",centered=True)
	metal_min_dim = max(pdk.get_grule("met2")["min_width"],pdk.get_grule("met3")["min_width"])
	metal_space = max(pdk.get_grule("met2")["min_separation"],pdk.get_grule("met3")["min_separation"],metal_min_dim)
	gate_route_os = evaluate_bbox(viam2m3)[0] - fetL.ports["multiplier_0_gate_W"].width + metal_space
	min_spacing_y = metal_space + 2*gate_route_os
	min_spacing_y = min_spacing_y - 2*abs(fetL.ports["well_S"].center[1] - fetL.ports["multiplier_0_gate_S"].center[1])
	# TODO: fix spacing where you see +-0.5
	a_topl = (diffpair << fetL).movey(fetL.ymax+min_spacing_y/2+0.5).movex(0-fetL.xmax-min_spacing_x/2)
	b_topr = (diffpair << fetR).movey(fetR.ymax+min_spacing_y/2+0.5).movex(fetL.xmax+min_spacing_x/2)
	a_botr = (diffpair << fetR)
	a_botr.mirror_y().movey(0-0.5-fetL.ymax-min_spacing_y/2).movex(fetL.xmax+min_spacing_x/2)
	b_botl = (diffpair << fetL)
	b_botl.mirror_y().movey(0-0.5-fetR.ymax-min_spacing_y/2).movex(0-fetL.xmax-min_spacing_x/2)
	# if substrate tap place substrate tap
	if substrate_tap:
		tapref = diffpair << tapring(pdk,evaluate_bbox(diffpair,padding=1),horizontal_glayer="met1")
		diffpair.add_ports(tapref.get_ports_list(),prefix="tap_")
		try:
			diffpair<<straight_route(pdk,a_topl.ports["multiplier_0_dummy_L_gsdcon_top_met_W"],diffpair.ports["tap_W_top_met_W"],glayer2="met1")
		except KeyError:
			pass
		try:
			diffpair<<straight_route(pdk,b_topr.ports["multiplier_0_dummy_R_gsdcon_top_met_W"],diffpair.ports["tap_E_top_met_E"],glayer2="met1")
		except KeyError:
			pass
		try:
			diffpair<<straight_route(pdk,b_botl.ports["multiplier_0_dummy_L_gsdcon_top_met_W"],diffpair.ports["tap_W_top_met_W"],glayer2="met1")
		except KeyError:
			pass
		try:
			diffpair<<straight_route(pdk,a_botr.ports["multiplier_0_dummy_R_gsdcon_top_met_W"],diffpair.ports["tap_E_top_met_E"],glayer2="met1")
		except KeyError:
			pass
	# route sources (short sources)
	diffpair << route_quad(a_topl.ports["multiplier_0_source_E"], b_topr.ports["multiplier_0_source_W"], layer=pdk.get_glayer("met2"))
	diffpair << route_quad(b_botl.ports["multiplier_0_source_E"], a_botr.ports["multiplier_0_source_W"], layer=pdk.get_glayer("met2"))
	sextension = b_topr.ports["well_E"].center[0] - b_topr.ports["multiplier_0_source_E"].center[0]
	source_routeE = diffpair << c_route(pdk, b_topr.ports["multiplier_0_source_E"], a_botr.ports["multiplier_0_source_E"],extension=sextension)
	source_routeW = diffpair << c_route(pdk, a_topl.ports["multiplier_0_source_W"], b_botl.ports["multiplier_0_source_W"],extension=sextension)
	# route drains
	# place via at the drain
	drain_br_via = diffpair << viam2m3
	drain_bl_via = diffpair << viam2m3
	drain_br_via.move(a_botr.ports["multiplier_0_drain_N"].center).movey(viam2m3.ymin)
	drain_bl_via.move(b_botl.ports["multiplier_0_drain_N"].center).movey(viam2m3.ymin)
	drain_br_viatm = diffpair << viam2m3
	drain_bl_viatm = diffpair << viam2m3
	drain_br_viatm.move(a_botr.ports["multiplier_0_drain_N"].center).movey(viam2m3.ymin)
	drain_bl_viatm.move(b_botl.ports["multiplier_0_drain_N"].center).movey(-1.5 * evaluate_bbox(viam2m3)[1] - metal_space)
	# create route to drain via
	width_drain_route = b_topr.ports["multiplier_0_drain_E"].width
	dextension = source_routeE.xmax - b_topr.ports["multiplier_0_drain_E"].center[0] + metal_space
	bottom_extension = viam2m3.ymax + width_drain_route/2 + 2*metal_space
	drain_br_viatm.movey(0-bottom_extension - metal_space - width_drain_route/2 - viam2m3.ymax)
	diffpair << route_quad(drain_br_viatm.ports["top_met_N"], drain_br_via.ports["top_met_S"], layer=pdk.get_glayer("met3"))
	diffpair << route_quad(drain_bl_viatm.ports["top_met_N"], drain_bl_via.ports["top_met_S"], layer=pdk.get_glayer("met3"))
	floating_port_drain_bottom_L = set_port_orientation(movey(drain_bl_via.ports["bottom_met_W"],0-bottom_extension), get_orientation("E"))
	floating_port_drain_bottom_R = set_port_orientation(movey(drain_br_via.ports["bottom_met_E"],0-bottom_extension - metal_space - width_drain_route), get_orientation("W"))
	drain_routeTR_BL = diffpair << c_route(pdk, floating_port_drain_bottom_L, b_topr.ports["multiplier_0_drain_E"],extension=dextension, width1=width_drain_route,width2=width_drain_route)
	drain_routeTL_BR = diffpair << c_route(pdk, floating_port_drain_bottom_R, a_topl.ports["multiplier_0_drain_W"],extension=dextension, width1=width_drain_route,width2=width_drain_route)
	# cross gate route top with c_route. bar_minus ABOVE bar_plus
	get_left_extension = lambda bar, a_topl=a_topl, diffpair=diffpair, pdk=pdk : (abs(diffpair.xmin-min(a_topl.ports["multiplier_0_gate_W"].center[0],bar.ports["e1"].center[0])) + pdk.get_grule("met2")["min_separation"])
	get_right_extension = lambda bar, b_topr=b_topr, diffpair=diffpair, pdk=pdk : (abs(diffpair.xmax-max(b_topr.ports["multiplier_0_gate_E"].center[0],bar.ports["e3"].center[0])) + pdk.get_grule("met2")["min_separation"])
	# lay bar plus and PLUSgate_routeW
	bar_comp = rectangle(centered=True,size=(abs(b_topr.xmax-a_topl.xmin), b_topr.ports["multiplier_0_gate_E"].width),layer=pdk.get_glayer("met2"))
	bar_plus = (diffpair << bar_comp).movey(diffpair.ymax + bar_comp.ymax + pdk.get_grule("met2")["min_separation"])
	PLUSgate_routeW = diffpair << c_route(pdk, a_topl.ports["multiplier_0_gate_W"], bar_plus.ports["e1"], extension=get_left_extension(bar_plus))
	#lay bar minus and MINUSgate_routeE
	plus_minus_seperation = max(pdk.get_grule("met2")["min_separation"], plus_minus_seperation)
	bar_minus = (diffpair << bar_comp).movey(diffpair.ymax +bar_comp.ymax + plus_minus_seperation)
	MINUSgate_routeE = diffpair << c_route(pdk, b_topr.ports["multiplier_0_gate_E"], bar_minus.ports["e3"], extension=get_right_extension(bar_minus))
	# lay MINUSgate_routeW and PLUSgate_routeE
	MINUSgate_routeW = diffpair << c_route(pdk, set_port_orientation(b_botl.ports["multiplier_0_gate_E"],"W"), bar_minus.ports["e1"], extension=get_left_extension(bar_minus))
	PLUSgate_routeE = diffpair << c_route(pdk, set_port_orientation(a_botr.ports["multiplier_0_gate_W"],"E"), bar_plus.ports["e3"], extension=get_right_extension(bar_plus))
	# correct pwell place, add ports, flatten, and return
	diffpair.add_ports(a_topl.get_ports_list(),prefix="tl_")
	diffpair.add_ports(b_topr.get_ports_list(),prefix="tr_")
	diffpair.add_ports(b_botl.get_ports_list(),prefix="bl_")
	diffpair.add_ports(a_botr.get_ports_list(),prefix="br_")
	diffpair.add_ports(source_routeE.get_ports_list(),prefix="source_routeE_")
	diffpair.add_ports(source_routeW.get_ports_list(),prefix="source_routeW_")
	diffpair.add_ports(drain_routeTR_BL.get_ports_list(),prefix="drain_routeTR_BL_")
	diffpair.add_ports(drain_routeTL_BR.get_ports_list(),prefix="drain_routeTL_BR_")
	diffpair.add_ports(MINUSgate_routeW.get_ports_list(),prefix="MINUSgateroute_W_")
	diffpair.add_ports(MINUSgate_routeE.get_ports_list(),prefix="MINUSgateroute_E_")
	diffpair.add_ports(PLUSgate_routeW.get_ports_list(),prefix="PLUSgateroute_W_")
	diffpair.add_ports(PLUSgate_routeE.get_ports_list(),prefix="PLUSgateroute_E_")
	diffpair.add_padding(layers=(pdk.get_glayer(well),), default=0)

	component = component_snap_to_grid(rename_ports_by_orientation(diffpair))

	component.info['netlist'] = diff_pair_netlist(fetL, fetR)
	return component







from os.path import join, dirname
from typing import Union, Optional
from copy import deepcopy

import re

class Netlist:
	"""Represents a SPICE netlist/subcircuit."""

	circuit_name: str = ""
	"""Name of the subcircuit."""
	nodes: list[str] = []
	"""List of nodes in the subcircuit."""
	global_nodes: list[str] = []
	"""List of the global nodes in the subcircuit."""

	designs_dir: str = join(dirname(__file__), 'designs')
	"""Path to the src/designs/ directory."""

	source_netlist: str = ""
	"""The source SPICE subcircuit (template) for the subcircuit. [Mako](https://makotemplates.org) templating syntax is supported."""
	instance_format: str = "X{name} {nodes} {circuit_name}"
	"""The format for the SPICE instantiation. Uses the Python string formatting syntax. `self.parameters`, `nodes`, `name`, and `circuit_name` are available parameters."""
	spice_netlist: str = ""
	"""The generated SPICE netlist."""

	sub_netlists: list['Netlist']
	"""List of the sub-netlists."""
	netlist_connections: list[list[str]]
	"""2D matrix of interconnections of the sub-netlists.

	The row and column number in the matrix represent the indices of the connected sub-netlists. The value represents the name of the wire connecting the nodes.
	"""

	# Variable to determine how many wires were created
	wire_index: int = 0
	"""The index of the next interconnection wire.

	The wires for interconnecting sub-netlists are named `wire{wire_index}`. This index is incremented each time a sub-netlist connection is made.
	"""

	parameters: dict = {}
	"""Dictionary of the high-level parameters."""

	def __init__(self, source_netlist: str = '', nodes: list[str] = [], circuit_name: Union[str, None] = None, instance_format: Optional[str] = None, parameters: dict = {}, sub_netlists: list['Netlist'] = []):
		"""Initializes a Netlist object.

		Override to load sub-netlists and parameters on initialization.
		"""
		self.parameters = {**self.parameters, **parameters}

		self.sub_netlists = []
		self.netlist_connections = []
		self.source_netlist = source_netlist
		self.nodes = nodes

		if circuit_name != None:
			self.circuit_name = circuit_name
		else:
			self.circuit_name = self.extract_subckt_name(self.source_netlist)

		if instance_format != None:
			self.instance_format = instance_format

		self.add_netlists(sub_netlists)

	def extract_subckt_name(self, netlist: str) -> str:
		"""Extracts the subcircuit name from the source SPICE."""
		for line in netlist.split('\n'):
			if line.count('subckt') != 0:
				return line.split(' ')[1]

		return 'Netlist'

	def generate_instance(self, name: Optional[str] = None, nodes: Optional[list[str]] = None, instance_format: Optional[str] = None) -> str:
		"""Generates an instance of the netlist subcircuit.
		Override to insert parameters in the instance.
		"""
		if name == None:
			name = self.circuit_name

		if instance_format == None:
			instance_format = self.instance_format

		if nodes == None:
			nodes = self.nodes

		params = {
			**self.generate_source_netlist_params(),
			'nodes': ' '.join(nodes),
			'name': name
		}

		return instance_format.format(**params)

	def read_source_netlist(self, netlist_src: str) -> str:
		"""Reads a source SPICE subcircuit template from a SPICE file. [Mako](https://makotemplates.org) templating syntax is supported in the source SPICE netlist."""
		self.source_netlist = open(join(self.designs_dir, netlist_src)).read()
		return self.source_netlist

	def connect_subnets(
		self,
		net1: Union[int, 'Netlist'],
		net2: Union[int, 'Netlist'],
		node_mapping: list[tuple[str, str]]
	):
		"""Adds a connection between two sub-netlists.

		Parameters:
			- `net1`: The netlist to connect. Either a reference to the Netlist object or it's index in the `sub_netlists` list.
			- `net2`: The netlist to connect to. Either a reference to the Netlist object or it's index in the `sub_netlists` list.
			- `node_mapping`: A list of 2-element tuples representing the connections between nodes of the netlists. The first element in the tuple is the name of the node of `net1` and the second value is the name of the node in `net2` to connect to.
		"""
		for mapping in node_mapping:
			node1, node2 = mapping

			net1_index = net1 if type(net1) == int else self.sub_netlists.index(net1)
			net2_index = net2 if type(net2) == int else self.sub_netlists.index(net2)

			netlist1 = self.sub_netlists[net1_index]
			netlist2 = self.sub_netlists[net2_index]

			node1_index = netlist1.nodes.index(node1)
			node2_index = netlist2.nodes.index(node2)

			connection_wire = ""
			# if one of the nodes is already connected, then use that wire instead of creating a new wire
			if re.match("^wire[\d]+$", self.netlist_connections[net1_index][node1_index]):
				connection_wire = self.netlist_connections[net1_index][node1_index]

			elif re.match("^wire[\d]+$", self.netlist_connections[net2_index][node2_index]):
				connection_wire = self.netlist_connections[net2_index][node2_index]

			else:
				connection_wire = f"wire{self.wire_index}"
				self.wire_index += 1

			self.netlist_connections[net1_index][node1_index] = connection_wire
			self.netlist_connections[net2_index][node2_index] = connection_wire


	def connect_node(
		self,
		net: Union[int, 'Netlist'],
		node_mapping: list[tuple[str, str]]
	):
		"""Connects a sub-netlist to a top-level node.

		Parameters:
		- `net`: The sub-netlist to connect. Either a reference to the Netlist object or it's index in the `sub_netlists` list.
		- `node_mapping`: A list of 2-element tuples representing the connections between the netlist nodes and the top-level nodes. The first element in the tuple is the name of the node of `net` and the second value is the name of the top-level to connect to.
		"""
		net_index = net if type(net) == int else self.sub_netlists.index(net)

		for mapping in node_mapping:
			net_node, top_level_node = mapping

			net = self.sub_netlists[net_index]
			node_index = net.nodes.index(net_node)

			self.netlist_connections[net_index][node_index] = top_level_node

	def add_netlists(self, netlists: list['Netlist']):
		"""Adds sub-netlists.

		Parameters:
		- `netlists`: A list of Netlist objects to add.
		"""
		for netlist in netlists:
			self.sub_netlists.append(netlist)
			self.netlist_connections.append(netlist.nodes.copy())

	def connect_netlist(self, netlist: 'Netlist', node_mapping: list[tuple[str, str]]) -> int:
		"""Adds a sub-netlist and connects it to top-level nodes.

		Parameters:
		- `netlist`: The netlist object to add.
		- `node_mapping`: A list of 2-element tuples representing the connections between the netlist nodes and the top-level nodes. The first element in the tuple is the name of the node of `netlist` and the second value is the name of the top-level to connect to.
		"""
		self.add_netlists([deepcopy(netlist)])
		netlist_index = len(self.sub_netlists) - 1

		self.connect_node(net=netlist_index, node_mapping=node_mapping)

		return netlist_index

	def generate_source_netlist_params(self, circuit_name: Optional[str] = None) -> dict:
		"""Generates the parameters to be inserted in the source SPICE netlist. Uses the Python template string format."""
		return {
			'circuit_name': circuit_name if circuit_name != None else self.circuit_name,
			'nodes': ' '.join(self.nodes),
			**self.parameters
		}

	def __generate_self_subcircuit(self, prefix: str = '', suffix: str = '') -> str:
		"""Generates the top-level SPICE subcircuit directive.
		The name of the subcircuit is set by `self.circuit_name`.
		"""
		generated_circuit_name = f"{prefix}{self.circuit_name}{suffix}"

		if self.source_netlist != "":
			return self.source_netlist.format(**self.generate_source_netlist_params(generated_circuit_name))

		elif len(self.sub_netlists) > 0:
			main_circuit = f".subckt {generated_circuit_name} {' '.join(self.nodes)}\n"

			for i, netlist in enumerate(self.sub_netlists):
				main_circuit += netlist.generate_instance(str(i), self.netlist_connections[i]) + "\n"

			main_circuit += f".ends {generated_circuit_name}"

			return main_circuit

		else:
			return ""

	def get_subcircuits_netlist_map(self, sub_netlists_only = False) -> dict[str, list['Netlist']]:
		"""Generates a list of all the unique SPICE subcircuits directives used in the netlist."""
		subcircuits = dict()

		# Get sub netlists' subcircuits
		for netlist in self.sub_netlists:
			subnetlist_subcircuits = netlist.get_subcircuits_netlist_map()

			# For each subnetlists' subcircuit
			for subckt in subnetlist_subcircuits:
				subckt_netlists = subnetlist_subcircuits[subckt]

				if subckt not in subcircuits:
					# If it is unique, create an entry in the dict
					subcircuits[subckt] = [*subckt_netlists]
				else:
					# If it is not unique, append the netlists to the dict
					subcircuits[subckt] += subckt_netlists


		if not sub_netlists_only:
			self_subckt = self.__generate_self_subcircuit()
			if self_subckt not in subcircuits:
				subcircuits[self_subckt] = [self]

		return subcircuits

	def get_global_nodes_list(self) -> set[str]:
		"""Generates a list of unique global nodes used in the netlist."""
		global_nodes = set()

		for netlist in self.sub_netlists:
			global_nodes.update(netlist.global_nodes)

		global_nodes.update(set(self.global_nodes))

		return global_nodes

	def generate_netlist(self, only_subcircuits: bool = False):
		"""Generates the final SPICE netlist for the design.

		The final netlist is a set of SPICE subcircuit directives and global directives. The top-level subcircuit is set by `self.circuit_name`.

		Parameters:
		- `only_subcircuits`: Only generates the subcircuit directives if set to `True`. (Default: `False`)
		"""

		# GENERATE UNIQUE SUBCIRCUIT DiRECTIVES
		subcircuits_netlist_map = self.get_subcircuits_netlist_map(sub_netlists_only=True)

		subcircuit_suffixes = dict()
		# Get the unique netlists' list and set their suffixes
		for subckt in subcircuits_netlist_map:
			netlists = subcircuits_netlist_map[subckt]
			# All of these subcircuits will have the same name, so use any one
			subckt_name = netlists[0].circuit_name

			if subckt_name in subcircuit_suffixes:
				# If a suffix exists, use it and increment it
				for netlist in netlists: netlist.circuit_name = f"{netlist.circuit_name}_{subcircuit_suffixes[subckt_name]}"
				subcircuit_suffixes[subckt_name] += 1
			else:
				# If a suffix doesn't exist, create it.
				subcircuit_suffixes[subckt_name] = 1

		unique_subcircuits_list = []
		# Generate all the unique subcircuits (generation is done again since suffixes were updated)
		for subckt in subcircuits_netlist_map:
			unique_subcircuits_list.append(
				# Use any one since all will be equal
				subcircuits_netlist_map[subckt][0].__generate_self_subcircuit()
			)
		# /GENERATE UNIQUE SUBCIRCUIT DiRECTIVES

		# GENERATE THE FINAL NETLIST
		subcircuits = '\n\n'.join(unique_subcircuits_list)
		main_circuit = self.__generate_self_subcircuit()
		global_nodes = ' '.join(self.get_global_nodes_list())

		self.spice_netlist = ""

		if len(global_nodes) > 0 and not only_subcircuits: self.spice_netlist += f".global {global_nodes}\n\n"

		self.spice_netlist += subcircuits + "\n\n"
		self.spice_netlist += main_circuit
		# /GENERATE THE FINAL NETLIST

		return self.spice_netlist



from gdsfactory.cell import cell, clear_cache
from gdsfactory.component import Component, copy
from gdsfactory.component_reference import ComponentReference
from gdsfactory.components.rectangle import rectangle
from glayout.pdk.mappedpdk import MappedPDK
from typing import Optional, Union
from glayout.primitives.fet import nmos, pmos, multiplier
from glayout.components.diff_pair import diff_pair
from glayout.primitives.guardring import tapring
from glayout.primitives.mimcap import mimcap_array, mimcap
from glayout.routing.L_route import L_route
from glayout.routing.c_route import c_route
from glayout.primitives.via_gen import via_stack, via_array
from gdsfactory.routing.route_quad import route_quad
from glayout.pdk.util.comp_utils import evaluate_bbox, prec_ref_center, movex, movey, to_decimal, to_float, move, align_comp_to_port, get_padding_points_cc
from glayout.pdk.util.port_utils import rename_ports_by_orientation, rename_ports_by_list, add_ports_perimeter, print_ports, set_port_orientation, rename_component_ports
from glayout.routing.straight_route import straight_route
from glayout.pdk.util.snap_to_grid import component_snap_to_grid
from pydantic import validate_arguments
from glayout.placement.two_transistor_interdigitized import two_nfet_interdigitized

from glayout.components.diff_pair_cmirrorbias import diff_pair_ibias
from glayout.components.stacked_current_mirror import stacked_nfet_current_mirror
from glayout.components.differential_to_single_ended_converter import differential_to_single_ended_converter
from glayout.components.row_csamplifier_diff_to_single_ended_converter import row_csamplifier_diff_to_single_ended_converter
from glayout.components.diff_pair_stackedcmirror import diff_pair_stackedcmirror
from glayout.spice import Netlist
from glayout.components.stacked_current_mirror import current_mirror_netlist

@validate_arguments
def __create_and_route_pins(
    pdk: MappedPDK,
    opamp_top: Component,
    pmos_comps_ref: ComponentReference,
    halfmultn_drain_routeref: ComponentReference,
    halfmultn_gate_routeref: ComponentReference
) -> tuple:
    _max_metal_seperation_ps = pdk.util_max_metal_seperation()
    # route halfmultp source, drain, and gate together, place vdd pin in the middle
    halfmultp_Lsrcport = opamp_top.ports["commonsource_Pamp_L_multiplier_0_source_con_N"]
    halfmultp_Rsrcport = opamp_top.ports["commonsource_Pamp_R_multiplier_0_source_con_N"]
    opamp_top << c_route(pdk, halfmultp_Lsrcport, halfmultp_Rsrcport, extension=opamp_top.ymax-halfmultp_Lsrcport.center[1], fullbottom=True,viaoffset=(False,False))
    # place vdd pin
    vddpin = opamp_top << rectangle(size=(5,3),layer=pdk.get_glayer("met4"),centered=True)
    vddpin.movey(opamp_top.ymax)
    # route vdd to source of 2L/2R
    opamp_top << straight_route(pdk, opamp_top.ports["pcomps_2L2Rsrcvia_top_met_N"], vddpin.ports["e4"])
    # drain route above vdd pin
    halfmultp_Ldrainport = opamp_top.ports["commonsource_Pamp_L_multiplier_0_drain_con_N"]
    halfmultp_Rdrainport = opamp_top.ports["commonsource_Pamp_R_multiplier_0_drain_con_N"]
    halfmultp_drain_routeref = opamp_top << c_route(pdk, halfmultp_Ldrainport, halfmultp_Rdrainport, extension=opamp_top.ymax-halfmultp_Ldrainport.center[1]+pdk.get_grule("met5")["min_separation"], fullbottom=True)
    halfmultp_Lgateport = opamp_top.ports["commonsource_Pamp_L_multiplier_0_gate_con_S"]
    halfmultp_Rgateport = opamp_top.ports["commonsource_Pamp_R_multiplier_0_gate_con_S"]
    ptop_halfmultp_gate_route = opamp_top << c_route(pdk, halfmultp_Lgateport, halfmultp_Rgateport, extension=abs(pmos_comps_ref.ymin-halfmultp_Lgateport.center[1])+pdk.get_grule("met5")["min_separation"],fullbottom=True,viaoffset=(False,False))
    # halfmultn to halfmultp drain to drain route
    extensionL = min(halfmultn_drain_routeref.ports["con_W"].center[0],halfmultp_drain_routeref.ports["con_W"].center[0])
    extensionR = max(halfmultn_drain_routeref.ports["con_E"].center[0],halfmultp_drain_routeref.ports["con_E"].center[0])
    opamp_top << c_route(pdk, halfmultn_drain_routeref.ports["con_W"], halfmultp_drain_routeref.ports["con_W"],extension=abs(opamp_top.xmin-extensionL)+2,cwidth=2)
    n_to_p_output_route = opamp_top << c_route(pdk, halfmultn_drain_routeref.ports["con_E"], halfmultp_drain_routeref.ports["con_E"],extension=abs(opamp_top.xmax-extensionR)+2,cwidth=2)
    # top nwell taps to vdd, top p substrate taps to gnd
    opamp_top << straight_route(pdk, opamp_top.ports["commonsource_cmirror_output_L_tie_N_top_met_N"], opamp_top.ports["pcomps_top_ptap_S_top_met_S"], width=5)
    opamp_top << straight_route(pdk, opamp_top.ports["commonsource_cmirror_output_R_tie_N_top_met_N"], opamp_top.ports["pcomps_top_ptap_S_top_met_S"], width=5)
    L_toptapn_route = opamp_top.ports["commonsource_Pamp_L_tie_N_top_met_N"]
    R_toptapn_route = opamp_top.ports["commonsource_Pamp_R_tie_N_top_met_N"]
    opamp_top << straight_route(pdk, movex(vddpin.ports["e4"],destination=L_toptapn_route.center[0]), L_toptapn_route, glayer1="met3",fullbottom=True)
    opamp_top << straight_route(pdk, movex(vddpin.ports["e4"],destination=R_toptapn_route.center[0]), R_toptapn_route, glayer1="met3",fullbottom=True)
    # bias pins for first two stages
    vbias1 = opamp_top << rectangle(size=(5,3),layer=pdk.get_glayer("met3"),centered=True)
    vbias1.movey(opamp_top.ymin - _max_metal_seperation_ps - vbias1.ymax)
    opamp_top << straight_route(pdk, vbias1.ports["e2"], opamp_top.ports["diffpair_ibias_B_gate_S"],width=1,fullbottom=False)
    vbias2 = opamp_top << rectangle(size=(5,3),layer=pdk.get_glayer("met5"),centered=True)
    vbias2.movex(1+opamp_top.xmax+evaluate_bbox(vbias2)[0]+pdk.util_max_metal_seperation()).movey(opamp_top.ymin+vbias2.ymax)
    opamp_top << L_route(pdk, halfmultn_gate_routeref.ports["con_E"], vbias2.ports["e2"],hwidth=2)
    # route + and - pins (being careful about antenna violations)
    minusi_pin = opamp_top << rectangle(size=(5,2),layer=pdk.get_glayer("met3"),centered=True)
    minusi_pin.movex(opamp_top.xmin).movey(_max_metal_seperation_ps + minusi_pin.ymax + halfmultn_drain_routeref.ports["con_W"].center[1] + halfmultn_drain_routeref.ports["con_W"].width/2)
    iport_antenna1 = movex(minusi_pin.ports["e3"],destination=opamp_top.ports["diffpair_MINUSgateroute_W_con_N"].center[0]-9*_max_metal_seperation_ps)
    opamp_top << L_route(pdk, opamp_top.ports["diffpair_MINUSgateroute_W_con_N"],iport_antenna1)
    iport_antenna2 = movex(iport_antenna1,offsetx=-9*_max_metal_seperation_ps)
    opamp_top << straight_route(pdk, iport_antenna1, iport_antenna2,glayer1="met4",glayer2="met4",via2_alignment=('c','c'),via1_alignment=('c','c'),fullbottom=True)
    iport_antenna2.layer=pdk.get_glayer("met4")
    opamp_top << straight_route(pdk, iport_antenna2, minusi_pin.ports["e3"],glayer1="met3",via2_alignment=('c','c'),via1_alignment=('c','c'),fullbottom=True)
    plusi_pin = opamp_top << rectangle(size=(5,2),layer=pdk.get_glayer("met3"),centered=True)
    plusi_pin.movex(opamp_top.xmin + plusi_pin.xmax).movey(_max_metal_seperation_ps + minusi_pin.ymax + plusi_pin.ymax)
    iport_antenna1 = movex(plusi_pin.ports["e3"],destination=opamp_top.ports["diffpair_PLUSgateroute_E_con_N"].center[0]-9*_max_metal_seperation_ps)
    opamp_top << L_route(pdk, opamp_top.ports["diffpair_PLUSgateroute_E_con_N"],iport_antenna1)
    iport_antenna2 = movex(iport_antenna1,offsetx=-9*_max_metal_seperation_ps)
    opamp_top << straight_route(pdk, iport_antenna1, iport_antenna2, glayer1="met4",glayer2="met4",via2_alignment=('c','c'),via1_alignment=('c','c'),fullbottom=True)
    iport_antenna2.layer=pdk.get_glayer("met4")
    opamp_top << straight_route(pdk, iport_antenna2, plusi_pin.ports["e3"],glayer1="met3",via2_alignment=('c','c'),via1_alignment=('c','c'),fullbottom=True)
    # route top center components to diffpair
    opamp_top << straight_route(pdk,opamp_top.ports["diffpair_tr_multiplier_0_drain_N"], opamp_top.ports["pcomps_pbottomAB_R_gate_S"], glayer1="met5",width=3*pdk.get_grule("met5")["min_width"],via1_alignment_layer="met2",via1_alignment=('c','c'))
    opamp_top << straight_route(pdk,opamp_top.ports["diffpair_tl_multiplier_0_drain_N"], opamp_top.ports["pcomps_minusvia_top_met_S"], glayer1="met5",width=3*pdk.get_grule("met5")["min_width"],via1_alignment_layer="met2",via1_alignment=('c','c'))
    # route minus transistor drain to output
    outputvia_diff_pcomps = opamp_top << via_stack(pdk,"met5","met4")
    outputvia_diff_pcomps.movex(opamp_top.ports["diffpair_tl_multiplier_0_drain_N"].center[0]).movey(ptop_halfmultp_gate_route.ports["con_E"].center[1])
    # add pin ports
    opamp_top.add_ports(vddpin.get_ports_list(), prefix="pin_vdd_")
    opamp_top.add_ports(vbias1.get_ports_list(), prefix="pin_diffpairibias_")
    opamp_top.add_ports(vbias2.get_ports_list(), prefix="pin_commonsourceibias_")
    opamp_top.add_ports(minusi_pin.get_ports_list(), prefix="pin_minus_")
    opamp_top.add_ports(plusi_pin.get_ports_list(), prefix="pin_plus_")
    #opamp_top.add_ports(output.get_ports_list(), prefix="pin_output_")
    return opamp_top, n_to_p_output_route



@validate_arguments
def __add_mimcap_arr(pdk: MappedPDK, opamp_top: Component, mim_cap_size, mim_cap_rows, ymin: float, n_to_p_output_route) -> tuple[Component, Netlist]:
    mim_cap_size = pdk.snap_to_2xgrid(mim_cap_size, return_type="float")
    max_metalsep = pdk.util_max_metal_seperation()
    mimcaps_ref = opamp_top << mimcap_array(pdk,mim_cap_rows,2,size=mim_cap_size,rmult=6)

    mimcap_netlist = mimcaps_ref.info['netlist']

    displace_fact = max(max_metalsep,pdk.get_grule("capmet")["min_separation"])
    mimcaps_ref.movex(pdk.snap_to_2xgrid(opamp_top.xmax + displace_fact + mim_cap_size[0]/2))
    mimcaps_ref.movey(pdk.snap_to_2xgrid(ymin + mim_cap_size[1]/2))
    # connect mimcap to gnd
    port1 = opamp_top.ports["pcomps_mimcap_connection_con_N"]
    port2 = mimcaps_ref.ports["row"+str(int(mim_cap_rows)-1)+"_col0_bottom_met_N"]
    cref2_extension = max_metalsep + opamp_top.ymax - max(port1.center[1], port2.center[1])
    opamp_top << c_route(pdk,port1,port2, extension=cref2_extension, fullbottom=True)
    intermediate_output = set_port_orientation(n_to_p_output_route.ports["con_S"],"E")
    opamp_top << L_route(pdk, mimcaps_ref.ports["row0_col0_top_met_S"], intermediate_output, hwidth=3)
    opamp_top.add_ports(mimcaps_ref.get_ports_list(),prefix="mimcap_")
    # add the cs output as a port
    opamp_top.add_port(name="commonsource_output_E", port=intermediate_output)
    return opamp_top, mimcap_netlist

def opamp_gain_stage_netlist(mimcap_netlist: Netlist, diff_cs_netlist: Netlist, cs_bias_netlist: Netlist) -> Netlist:
    netlist = Netlist(
        circuit_name="GAIN_STAGE",
        nodes=['VIN1', 'VIN2', 'VOUT', 'VDD', 'IBIAS', 'GND']
    )

    diff_cs_ref = netlist.connect_netlist(
        diff_cs_netlist,
        [('VSS', 'VDD')]
    )

    netlist.connect_netlist(
        cs_bias_netlist,
        [('VREF', 'IBIAS'), ('VSS', 'GND'), ('VCOPY', 'VOUT')]
    )

    mimcap_ref = netlist.connect_netlist(mimcap_netlist, [('V2', 'VOUT')])

    netlist.connect_subnets(
        mimcap_ref,
        diff_cs_ref,
        [('V1', 'VSS2')]
    )

    return netlist

def opamp_twostage_netlist(input_stage_netlist: Netlist, gain_stage_netlist: Netlist) -> Netlist:
    two_stage_netlist = Netlist(
        circuit_name="OPAMP_TWO_STAGE",
        nodes=['VDD', 'GND', 'DIFFPAIR_BIAS', 'VP', 'VN', 'CS_BIAS', 'VOUT']
    )

    input_stage_ref = two_stage_netlist.connect_netlist(
        input_stage_netlist,
        [('IBIAS', 'DIFFPAIR_BIAS'), ('VSS', 'GND'), ('B', 'GND')]
    )

    gain_stage_ref = two_stage_netlist.connect_netlist(
        gain_stage_netlist,
        [('IBIAS', 'CS_BIAS')]
    )

    two_stage_netlist.connect_subnets(
        input_stage_ref,
        gain_stage_ref,
        [('VDD1', 'VIN1'), ('VDD2', 'VIN2')]
    )

    return two_stage_netlist

def opamp_twostage(
    pdk: MappedPDK,
    half_diffpair_params: tuple[float, float, int] = (6, 1, 4),
    diffpair_bias: tuple[float, float, int] = (6, 2, 4),
    half_common_source_params: tuple[float, float, int, int] = (7, 1, 10, 3),
    half_common_source_bias: tuple[float, float, int, int] = (6, 2, 8, 2),
    half_pload: tuple[float,float,int] = (6,1,6),
    mim_cap_size=(12, 12),
    mim_cap_rows=3,
    rmult: int = 2,
    with_antenna_diode_on_diffinputs: int=5
) -> Component:
    """
    create a two stage opamp, args->
    pdk: pdk to use
    half_diffpair_params: diffpair (width,length,fingers)
    diffpair_bias: bias transistor for diffpair nmos (width,length,fingers). The ref and output of the cmirror are identical
    half_common_source_params: pmos top component amp (width,length,fingers,mults)
    half_common_source_bias: bottom L/R large nmos current mirror (width,length,fingers,mults). The ref of the cmirror always has 1 multplier. multiplier must be >=2
    ****NOTE: change the multiplier option to change the relative sizing of the current mirror ref/output
    half_pload: all 4 pmos load transistors of first stage (width,length,...). The last element in the tuple is the fingers of the bottom two pmos.
    mim_cap_size: width,length of individual mim_cap
    mim_cap_rows: number of rows in the mimcap array (always 2 cols)
    rmult: routing multiplier (larger = wider routes)
    with_antenna_diode_on_diffinputs: adds antenna diodes with_antenna_diode_on_diffinputs*(1um/0.5um) on the positive and negative inputs to the opamp
    """
    # error checks
    if with_antenna_diode_on_diffinputs!=0 and with_antenna_diode_on_diffinputs<2:
        raise ValueError("number of antenna diodes should be at least 2 (or 0 to specify no diodes)")
    if half_common_source_bias[3] < 2:
        raise ValueError("half_common_source_bias num multiplier must be >= 2")
    opamp_top, halfmultn_drain_routeref, halfmultn_gate_routeref, _cref = diff_pair_stackedcmirror(pdk, half_diffpair_params, diffpair_bias, half_common_source_bias, rmult, with_antenna_diode_on_diffinputs)

    opamp_top.info['netlist'].circuit_name = "INPUT_STAGE"

    # place pmos components
    pmos_comps = differential_to_single_ended_converter(pdk, rmult, half_pload, opamp_top.ports["diffpair_tl_multiplier_0_drain_N"].center[0])
    clear_cache()

    pmos_comps = row_csamplifier_diff_to_single_ended_converter(pdk, pmos_comps, half_common_source_params, rmult)

    cs_bias_netlist = current_mirror_netlist(
        pdk,
        width=diffpair_bias[0],
        length=diffpair_bias[1],
        multipliers=diffpair_bias[2]
    )

    ydim_ncomps = opamp_top.ymax
    pmos_comps_ref = opamp_top << pmos_comps
    pmos_comps_ref.movey(round(ydim_ncomps + pmos_comps_ref.ymax+10))
    opamp_top.add_ports(pmos_comps_ref.get_ports_list(),prefix="pcomps_")
    rename_func = lambda name_, port_ : name_.replace("pcomps_halfpspecialmarker","commonsource_Pamp") if name_.startswith("pcomps_halfpspecialmarker") else name_
    opamp_top = rename_component_ports(opamp_top, rename_function=rename_func)
    # create pins and route
    clear_cache()
    opamp_top, n_to_p_output_route = __create_and_route_pins(pdk, opamp_top, pmos_comps_ref, halfmultn_drain_routeref, halfmultn_gate_routeref)
    # place mimcaps and route
    clear_cache()
    opamp_top, mimcap_netlist = __add_mimcap_arr(pdk, opamp_top, mim_cap_size, mim_cap_rows, pmos_comps_ref.ymin, n_to_p_output_route)
    opamp_top.add_ports(n_to_p_output_route.get_ports_list(),"special_con_npr_")
    # return
    opamp_top.add_ports(_cref.get_ports_list(), prefix="gnd_route_")

    pmos_comps.info['netlist'] = opamp_gain_stage_netlist(mimcap_netlist, pmos_comps.info['netlist'], cs_bias_netlist)
    opamp_top.info['netlist'] = opamp_twostage_netlist(opamp_top.info['netlist'], pmos_comps.info['netlist'])

    return opamp_top





from pydantic import validate_arguments
from gdsfactory.typings import Component, ComponentReference
from gdsfactory.components.rectangle import rectangle
from gdsfactory.port import Port
from typing import Callable, Union, Optional
from decimal import Decimal
from pathlib import Path
import pickle
from PrettyPrint import PrettyPrintTree


@validate_arguments
def rename_component_ports(custom_comp: Component, rename_function: Callable[[str, Port], str]) -> Component:
    """uses rename_function(str, Port) -> str to decide which ports to rename.
    rename_function accepts the current port name (string) and current port (Port) then returns the new port name
    rename_function can return new name = current port name, in which case the name will not change
    rename_function should raise error if custom requirments for rename are not met
    if you want to pass additional args to rename_function, implement a functor
    custom_comp is the components to modify. the modified component is returned
    """
    names_to_modify = list()
    # find ports and get new names
    for pname, pobj in custom_comp.ports.items():
        # error checking
        if not pname == pobj.name:
            raise ValueError("component may have an invalid ports dict")
        
        new_name = rename_function(pname, pobj)
        
        names_to_modify.append((pname,new_name))
    # modify names
    for namepair in names_to_modify:
        if namepair[0] in custom_comp.ports.keys():
            portobj = custom_comp.ports.pop(namepair[0])
            portobj.name = namepair[1]
            custom_comp.ports[namepair[1]] = portobj
        else:
            raise KeyError("name "+str(namepair[0])+" not in component ports")
    # returns modified component/component ref
    return custom_comp


@validate_arguments
def rename_ports_by_orientation__call(old_name: str, pobj: Port) -> str:
	"""internal implementation of port orientation rename"""
	if not "_" in old_name and not any(old_name==edge for edge in ["e1","e2","e3","e4"]):
		raise ValueError("portname must contain underscore \"_\" " + old_name)
	# get new suffix (port orientation)
	new_suffix = None
	angle = pobj.orientation % 360 if pobj.orientation is not None else 0
	angle = round(angle)
	if angle <= 45 or angle >= 315:
		new_suffix = "E"
	elif angle <= 135 and angle >= 45:
		new_suffix = "N"
	elif angle <= 225 and angle >= 135:
		new_suffix = "W"
	else:
		new_suffix = "S"
	# handle special case where no underscore and name is e1/2/3/4
	if any(old_name==edge for edge in ["e1","e2","e3","e4"]):
		return new_suffix
	# construct new name
	old_str_split = old_name.rsplit("_", 1)
	old_str_split[1] = new_suffix
	new_name = "_".join(old_str_split)
	return new_name

@validate_arguments
def rename_ports_by_orientation(custom_comp: Component) -> Component:
    """replaces the last part of the port name 
    (after the last underscore, unless name is e1/2/3/4) with a direction
    direction is one of N,E,S,W
    returns the modified component
    """
    return rename_component_ports(custom_comp, rename_ports_by_orientation__call)


class rename_ports_by_list__call: 
	def __init__(self, replace_list: list[tuple[str,str]] = []): 
		self.replace_list = dict(replace_list)
		self.replace_history = dict.fromkeys(self.replace_list.keys())
		for keyword in self.replace_history:
			self.replace_history[keyword] = 0
	@validate_arguments
	def __call__(self, old_name: str, pobj: Port) -> str:
		for keyword, newname in self.replace_list.items():
			if keyword in old_name:
				inst_id = self.replace_history[keyword]
				replace_name = newname + str(inst_id if inst_id else "")
				self.replace_history[keyword] += 1
				return replace_name
		return old_name

@validate_arguments
def rename_ports_by_list(custom_comp: Component, replace_list: list[tuple[str,str]]) -> Component:
    """replace_list is a list of tuple(string, string)
    if a port name contains tuple[0], the port will be renamed to tuple[1]
    if tuple[1] is None or empty string raise error
    when anaylzing a single port, if multiple keywords from the replace_list are found, first match is returned
    since we cannot have duplicate port names, different ports that end up with the same name get numbered"""
    rename_func = rename_ports_by_list__call(replace_list)
    return rename_component_ports(custom_comp, rename_func)


def remove_ports_with_prefix(custom_comp: Component, prefix: str) -> Component:
	"""remove all ports in custom_comp which begin with prefix"""
	# determine which ports to remove
	remove_list = list()
	for prt in custom_comp.ports.keys():
		if prt.startswith(prefix):
			remove_list.append(prt)
	# remove the ports
	for prt in remove_list:
		custom_comp.ports.pop(prt)
	return custom_comp


@validate_arguments
def add_ports_perimeter(custom_comp: Component, layer: tuple[int, int], prefix: Optional[str] = "_") -> Component:
	"""adds ports to the outside perimeter of a cell
	custom_comp = component to add ports to (returns the modified component)
	layer = will extract this layer and take it as the bbox, ports will also be on this layer
	prefix = prefix to add to the port names. Adds an underscore by default
	returns ports named by orientation
	"""
	if "_" not in prefix:
		raise ValueError("you need underscore char in prefix")
	compbbox = custom_comp.extract(layers=(layer,)).bbox
	width = compbbox[1][0] - compbbox[0][0]
	height = compbbox[1][1] - compbbox[0][1]
	custom_comp.add_port(name=prefix+"W",width=height,orientation=180,center=(compbbox[0][0],compbbox[0][1]+height/2),layer=layer,port_type="electrical")
	custom_comp.add_port(name=prefix+"N",width=width,orientation=90,center=(compbbox[0][0]+width/2,compbbox[1][1]),layer=layer,port_type="electrical")
	custom_comp.add_port(name=prefix+"E",width=height,orientation=0,center=(compbbox[1][0],compbbox[0][1]+height/2),layer=layer,port_type="electrical")
	custom_comp.add_port(name=prefix+"S",width=width,orientation=270,center=(compbbox[0][0]+width/2,compbbox[0][1]),layer=layer,port_type="electrical")
	return custom_comp


@validate_arguments
def get_orientation(orientation: Union[int,float,str], int_only: bool=False) -> Union[float,int,str]:
	"""returns the angle corresponding to port orientation
	orientation must contain N/n,E/e,S/s,W/w
	e.g. all the follwing are valid:
	N/n or N/north,E/e or E/east,S/s or S/south, W/w or W/west
	if int_only, will return int regardless of input type,
	else will return the opposite type of that given
	(i.e. will return str if given int/float and int if given str)
	"""
	if isinstance(orientation,str):
		orientation = orientation.lower()
		if "n" in orientation:
			return 90
		elif "e" in orientation:
			return 0
		elif "w" in orientation:
			return 180
		elif "s" in orientation:
			return 270
		else:
			raise ValueError("orientation must contain N/n,E/e,S/s,W/w")
	else:# must be a float/int
		orientation = int(orientation)
		if int_only:
			return orientation
		orientation_index = int((orientation % 360) / 90)
		orientations = ["E","N","W","S"]
		try:
			orientation = orientations[orientation_index]
		except IndexError as e:
			raise ValueError("orientation must be 0,90,180,270 to use this function")
		return orientation


@validate_arguments
def assert_port_manhattan(edges: Union[list[Port],Port]) -> bool:
	"""raises assertionerror if port is not vertical or horizontal"""
	if isinstance(edges, Port):
		edges = [edges]
	for edge in edges:
		if round(edge.orientation) % 90 != 0:
			raise AssertionError("edge is not vertical or horizontal")
	return True


@validate_arguments
def assert_ports_perpindicular(edge1: Port, edge2: Port) -> bool:
	"""raises assertionerror if edges are not perindicular"""
	or1 = round(edge1.orientation)
	or2 = round(edge2.orientation)
	angle_difference = abs(round(or1-or2))
	if angle_difference != 90 and angle_difference != 270:
		raise AssertionError("edges are not perpindicular")
	return True


@validate_arguments
def set_port_orientation(custom_comp: Port, orientation: Union[float, int, str], flip180: Optional[bool]=False) -> Port:
	"""creates a new port with the desired orientation and returns the new port"""
	if isinstance(orientation,str):
		orientation = get_orientation(orientation, int_only=True)
	if flip180:
		orientation = (orientation + 180) % 360
	newport = Port(
		name = custom_comp.name,
		center = custom_comp.center,
		orientation = orientation,
		parent = custom_comp.parent,
		port_type = custom_comp.port_type,
		cross_section = custom_comp.cross_section,
		shear_angle = custom_comp.shear_angle,
		layer = custom_comp.layer,
		width = custom_comp.width,
	)
	return newport


@validate_arguments
def set_port_width(custom_comp: Port, width: float) -> Port:
	"""creates a new port with the desired width and returns the new port"""
	newport = Port(
		name = custom_comp.name,
		center = custom_comp.center,
		orientation = custom_comp.orientation,
		parent = custom_comp.parent,
		port_type = custom_comp.port_type,
		cross_section = custom_comp.cross_section,
		shear_angle = custom_comp.shear_angle,
		layer = custom_comp.layer,
		width = width,
	)
	return newport


@validate_arguments
def print_ports(custom_comp: Union[Component, ComponentReference], names_only: Optional[bool] = True) -> None:
    """prints ports in comp in a nice way
    custom_comp = component to use
    names_only = only print names if True else print name and port
    """
    for key,val in custom_comp.ports.items():
        print(key)
        if not names_only:
            print(val)
            print()


class PortTree:
	"""PortTree helps a glayout programmer visualize the ports in a component
	\"_\" should represent a level of hiearchy (much like a directory). think of this like psuedo directories
	Initialize a PortTree from a Component or ComponentReference
	then use self.ls to list all ports/subdirectories in a directory
	you can use self.print to prettyprint a port tree (uses pypi prettyprinttree package)

	You should not need to access the internal dictionary for the tree, but if you do:
	PortTree internally uses tuple[str, dict] = name:children as the node type
	since the PortTree is not a node type (PortTree is not a real tree class), the root node is: (self.name, self.tree)
	"""

	@validate_arguments
	def __init__(self, custom_comp: Union[Component, ComponentReference], name: Optional[str]=None):
		"""creates the tree structure from the ports where _ represent subdirectories
		credit -> chatGPT
		"""
		file_list = custom_comp.ports.keys()
		directory_tree = {}
		for file_path in file_list:
			path_components = file_path.split('_')
			current_dir = directory_tree
			for path_component in path_components:
				if path_component not in current_dir:
					current_dir[path_component] = {}
				current_dir = current_dir[path_component]
		self.tree = directory_tree
		self.name = name if name else custom_comp.name
	
	@validate_arguments
	def ls(self, file_path: Optional[str] = None) -> list[str]:
		"""tries to traverse the tree along the given path and prints all subdirectories in a psuedo directory
		if the path given is not found in the tree, raises KeyError
		path should not end with \"_\" char
		"""
		if file_path is None or len(file_path)==0:
			return list(self.tree.keys())
		path_components = file_path.split('_')
		current_dir = self.tree
		for path_component in path_components:
			if path_component not in current_dir:
				raise KeyError("Port path was not found")
			current_dir = current_dir[path_component]
		return list(current_dir.keys())
	
	@validate_arguments
	def save_to_disk(self, savedir: Union[Path, str]="./"):
		savedir = Path(savedir).resolve()
		savedir.mkdir(exist_ok=True,parents=True)
		if not savedir.is_dir():
			raise ValueError("no dir named" + str(savedir))
		with open(savedir / "porttree.pkl", 'wb') as outfile:
			pickle.dump(self, outfile)
	
	@classmethod
	def read_from_disk(cls, pklfile: Union[Path, str]):
		pklfile = Path(pklfile).resolve()
		if not pklfile.is_file():
			raise ValueError("no file named" + str(pklfile))
		with open(str(pklfile), 'rb') as infile:
			return pickle.load(infile)
	
	def get_children(self, node: tuple[str, dict]) -> list[tuple[str, dict]]:
		"""access children of internal tree node (node might be a PortTree)"""
		node_dict = node[1] if isinstance(node, tuple) else self.tree
		return node_dict.items()
		
	
	def get_val(self, node: tuple[str, dict]) -> str:
		"""returns value of a node, (node might be a PortTree)"""
		return node[0] if isinstance(node, tuple) else self.name
	
	def print(self, savetofile: bool=True, default_opts: bool=True, depth: Optional[int]=None, outfile_name: Optional[str]=None, **kwargs):
		"""prints output to terminal directly using prettyprinttree pypi package
		args:
		depth = max depth to print. this is a kwarg but since it so common, it should be specfied from depth arg
		savetofile = saves print output to a txt file rather than printing to terminal (easier to view, but without nice formatting)
		default_opts = bool=True results in using glayout recommended default print arguments
		kwargs -> kwargs are prettyprint options passed directly to prettyprint.
		****NOTE: kwargs override all other options
		"""
		depth = int(depth) if (depth is not None and depth>0) else -1
		extra_kwargs = {}
		if default_opts:
			extra_kwargs.update({"default_orientation": True})
		if savetofile:
			extra_kwargs.update({"return_instead_of_print":savetofile, "color":None, "border":True, "default_orientation": True})
		extra_kwargs.update(kwargs)
		pt = PrettyPrintTree(self.get_children, self.get_val, max_depth=depth, **extra_kwargs)
		rtrstr = pt(self)
		if rtrstr:
			outfile_name = "outputtree.txt" if outfile_name is None else outfile_name
			with open(outfile_name,"w") as outputfile:
				outputfile.write(rtrstr)


def print_port_tree_all_cells() -> list:
	"""print the PortTree for most of the glayout cells and save as a text file.
	returns a list of components
	"""
	from glayout.primitives.via_gen import via_stack, via_array
	from glayout.opamp import opamp
	from glayout.primitives.mimcap import mimcap
	from glayout.primitives.mimcap import mimcap_array
	from glayout.primitives.guardring import tapring
	from glayout.primitives.fet import multiplier, nmos, pmos
	from glayout.diff_pair import diff_pair
	from glayout.routing.straight_route import straight_route
	from glayout.routing.c_route import c_route
	from glayout.routing.L_route import L_route
	from glayout.pdk.sky130_mapped import sky130_mapped_pdk as pdk
	from gdsfactory.port import Port
	print("saving via_stack, via_array, opamp, mimcap, mimcap_array, tapring, multiplier, nmos, pmos, diff_pair, straight_route, c_route, L_route Ports to txt files")
	celllist = list()
	celllist.append(["via_stack",via_stack(pdk, "active_diff","met5")])
	celllist.append(["viaarray", via_array(pdk, "active_diff","met5", num_vias=(2,3))])
	celllist.append(["mimcap",mimcap(pdk)])
	celllist.append(["mimcap_array",mimcap_array(pdk, 2, 3)])
	celllist.append(["tapring",tapring(pdk)])
	celllist.append(["multiplier",multiplier(pdk,"n+s/d")])
	celllist.append(["nmos", nmos(pdk,fingers=2,multipliers=2)])
	celllist.append(["pmos", pmos(pdk,fingers=2,multipliers=2)])
	celllist.append(["diff_pair",diff_pair(pdk)])
	psuedo_porta = Port("bottom",90,(0,0),2,layer=pdk.get_glayer("met2"))
	psuedo_portb = Port("top",0,(5,10),2.5,layer=pdk.get_glayer("met5"))
	psuedo_porta = Port("right",90,(10,10),2,layer=pdk.get_glayer("met2"))
	celllist.append(["straight_route",straight_route(pdk,psuedo_porta,psuedo_portb)])
	celllist.append(["L_route",L_route(pdk, psuedo_porta, psuedo_portb)])
	celllist.append(["c_route",c_route(pdk, psuedo_porta, psuedo_porta,extension=2)])
	celllist.append(["opamp",opamp(pdk)])
	for name, py_cell in celllist:
		from glayout import __version__ as glayoutvinfo
		glayoutv = str(glayoutvinfo)
		PortTree(py_cell,name=name).print(depth=5,outfile_name=name+"_v"+glayoutv+"_tree.txt",default_orientation=True)
	return celllist



from gdsfactory.cell import cell, clear_cache
from gdsfactory.component import Component, copy
from gdsfactory.component_reference import ComponentReference
from gdsfactory.components.rectangle import rectangle
from glayout.pdk.mappedpdk import MappedPDK
from typing import Optional, Union
from glayout.primitives.fet import nmos, pmos, multiplier
from glayout.components.diff_pair import diff_pair
from glayout.primitives.guardring import tapring
from glayout.primitives.mimcap import mimcap_array, mimcap
from glayout.routing.L_route import L_route
from glayout.routing.c_route import c_route
from glayout.primitives.via_gen import via_stack, via_array
from gdsfactory.routing.route_quad import route_quad
from glayout.pdk.util.comp_utils import evaluate_bbox, prec_ref_center, movex, movey, to_decimal, to_float, move, align_comp_to_port, get_padding_points_cc
from glayout.pdk.util.port_utils import rename_ports_by_orientation, rename_ports_by_list, add_ports_perimeter, print_ports, set_port_orientation, rename_component_ports
from glayout.routing.straight_route import straight_route
from glayout.pdk.util.snap_to_grid import component_snap_to_grid
from pydantic import validate_arguments
from glayout.placement.two_transistor_interdigitized import two_nfet_interdigitized
from glayout.spice import Netlist

def current_mirror_netlist(pdk: MappedPDK, width: float, length: float, multipliers: int) -> Netlist:
    return Netlist(
        circuit_name='CURRENT_MIRROR',
        nodes=['VREF', 'VCOPY', 'VSS'],
        source_netlist=""".subckt {circuit_name} {nodes} l=1 w=1 m=1
XREF VREF VREF VSS VSS {model} l={{l}} w={{w}} m={{m}}
XCOPY VCOPY VREF VSS VSS {model} l={{l}} w={{w}} m={{m}}
.ends {circuit_name}""",
        instance_format="X{name} {nodes} {circuit_name} l={length} w={width} m={mult}",
        parameters={
            'model': pdk.models['nfet'],
            'width': width,
            'length': length,
            'mult': multipliers
        }
    )

@validate_arguments
def stacked_nfet_current_mirror(pdk: MappedPDK, half_common_source_nbias: tuple[float, float, int, int], rmult: int, sd_route_left: bool) -> Component:
    cmirror_output = nmos(
        pdk,
        width=half_common_source_nbias[0],
        length=half_common_source_nbias[1],
        fingers=half_common_source_nbias[2],
        multipliers=half_common_source_nbias[3],
        with_tie=True,
        with_dnwell=False,
        with_substrate_tap=False,
        with_dummy=True,
        sd_route_left = sd_route_left,
        rmult=rmult,
        tie_layers=("met2","met2")
    )
    cmirrorref = nmos(
        pdk,
        width=half_common_source_nbias[0],
        length=half_common_source_nbias[1],
        fingers=half_common_source_nbias[2],
        multipliers=1,
        with_tie=True,
        with_dnwell=False,
        with_substrate_tap=False,
        with_dummy=True,
        sd_route_left = sd_route_left,
        rmult=rmult,
        tie_layers=("met2","met2")
    )
    cmirrorref_ref = prec_ref_center(cmirrorref)
    cmirrorout_ref = prec_ref_center(cmirror_output)
    return cmirrorref_ref, cmirrorout_ref




from setuptools import find_packages
from setuptools import setup


with open("glayout/README.md") as f:
    LONG_DESCRIPTION = f.read()


def get_install_requires():
    with open("glayout/requirements.txt", "r") as f:
        return [line.strip() for line in f.readlines() if not line.startswith("-")]


setup(
    name="glayout",
    version="0.0.3",
    url="https://github.com/idea-fasoc/OpenFASOC",
    license="MIT",
    author=" msaligane",
    author_email="mehdi@umich.edu",
    description="Fully Open-Source Autonomous SoC Synthesis using Customizable Cell-Based Synthesizable Analog Circuits",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=("tests",)),
    package_dir={"":"./"},
    package_data={'glayout' :[]},
    include_package_data=True,
    install_requires=get_install_requires(),
    python_requires=">=3.10",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.10",
    ],
)



from gdsfactory.cell import cell, clear_cache
from gdsfactory.component import Component, copy
from gdsfactory.component_reference import ComponentReference
from gdsfactory.components.rectangle import rectangle
from glayout.pdk.mappedpdk import MappedPDK
from typing import Optional, Union
from glayout.primitives.fet import nmos, pmos, multiplier
from glayout.components.diff_pair import diff_pair
from glayout.primitives.guardring import tapring
from glayout.primitives.mimcap import mimcap_array, mimcap
from glayout.routing.L_route import L_route
from glayout.routing.c_route import c_route
from glayout.primitives.via_gen import via_stack, via_array
from gdsfactory.routing.route_quad import route_quad
from glayout.pdk.util.comp_utils import evaluate_bbox, prec_ref_center, movex, movey, to_decimal, to_float, move, align_comp_to_port, get_padding_points_cc
from glayout.pdk.util.port_utils import rename_ports_by_orientation, rename_ports_by_list, add_ports_perimeter, print_ports, set_port_orientation, rename_component_ports
from glayout.routing.straight_route import straight_route
from glayout.pdk.util.snap_to_grid import component_snap_to_grid
from pydantic import validate_arguments
from glayout.placement.two_transistor_interdigitized import two_nfet_interdigitized
from glayout.spice import Netlist

from glayout.components.opamp_twostage import opamp_twostage
from glayout.components.stacked_current_mirror import current_mirror_netlist

def opamp_output_stage_netlist(pdk: MappedPDK, output_amp_fet_ref: ComponentReference, biasParams: list) -> Netlist:
    bias_netlist = current_mirror_netlist(pdk, biasParams[0], biasParams[1], biasParams[2])

    output_stage_netlist = Netlist(
        circuit_name="OUTPUT_STAGE",
        nodes=['VDD', 'GND', 'IBIAS', 'VIN', 'VOUT']
    )

    output_stage_netlist.connect_netlist(
        output_amp_fet_ref.info['netlist'],
        [('D', 'VDD'), ('G', 'VIN'), ('B', 'GND'), ('S', 'VOUT')]
    )

    output_stage_netlist.connect_netlist(
        bias_netlist,
        [('VREF', 'IBIAS'), ('VSS', 'GND'), ('VCOPY', 'VOUT')]
    )

    return output_stage_netlist

@validate_arguments
def __add_output_stage(
    pdk: MappedPDK,
    opamp_top: Component,
    amplifierParams: tuple[float, float, int],
    biasParams: list,
    rmult: int,
) -> tuple[Component, Netlist]:
    '''add output stage to opamp_top, args:
    pdk = pdk to use
    opamp_top = component to add output stage to
    amplifierParams = [width,length,fingers,mults] for amplifying FET
    biasParams = [width,length,fingers,mults] for bias FET
    '''
    # Instantiate output amplifier
    amp_fet_ref = opamp_top << nmos(
        pdk,
        width=amplifierParams[0],
        length=amplifierParams[1],
        fingers=amplifierParams[2],
        multipliers=1,
        sd_route_topmet="met3",
        gate_route_topmet="met3",
        rmult=rmult,
        with_dnwell=False,
        with_tie=True,
        with_substrate_tap=False,
        tie_layers=("met2","met2")
    )
    # Instantiate bias FET
    cmirror_ibias = opamp_top << two_nfet_interdigitized(
        pdk,
        numcols=biasParams[2],
        width=biasParams[0],
        length=biasParams[1],
        fingers=1,
        gate_route_topmet="met3",
        sd_route_topmet="met3",
        rmult=rmult,
        with_substrate_tap=False,
        tie_layers=("met2","met2")
    )

    metal_sep = pdk.util_max_metal_seperation()
    # Locate output stage relative position
    # x-coordinate: Center of SW capacitor in array
    # y-coordinate: Top of NMOS blocks
    xref_port = opamp_top.ports["mimcap_row0_col0_bottom_met_S"]
    x_cord = xref_port.center[0] - xref_port.width/2
    y_cord = opamp_top.ports["commonsource_cmirror_output_R_tie_N_top_met_N"].center[1]
    dims = evaluate_bbox(amp_fet_ref)
    center = [x_cord + dims[0]/2, y_cord - dims[1]/2]
    amp_fet_ref.move(center)
    amp_fet_ref.movey(pdk.get_grule("active_tap", "p+s/d")["min_enclosure"])
    dims = evaluate_bbox(cmirror_ibias)
    cmirror_ibias.movex(amp_fet_ref.xmin + dims[0]/2)
    cmirror_ibias.movey(amp_fet_ref.ymin - dims[1]/2 - metal_sep)
    # route input of output_stage to output of previous stage
    n_to_p_output_route = opamp_top.ports["special_con_npr_con_S"]
    opamp_top << L_route(pdk, n_to_p_output_route, amp_fet_ref.ports["multiplier_0_gate_W"])
    # route drain of amplifier to vdd
    vdd_route_extension = opamp_top.ymax-opamp_top.ports["pin_vdd_e4"].center[1]+metal_sep
    opamp_top << c_route(pdk,amp_fet_ref.ports["multiplier_0_drain_N"],set_port_orientation(opamp_top.ports["pin_vdd_e4"],"N"),width1=5,width2=5,extension=vdd_route_extension,e2glayer="met3")
    vddvia = opamp_top << via_stack(pdk,"met3","met4",fullbottom=True)
    align_comp_to_port(vddvia,opamp_top.ports["pin_vdd_e4"],('c','t'))
    # route drain of cmirror to source of amplifier
    opamp_top << c_route(pdk, cmirror_ibias.ports["B_drain_E"],amp_fet_ref.ports["multiplier_0_source_E"],extension=metal_sep)
    # route cmirror: A gate, B gate and A drain together. Then A source and B source to ground
    gate_short = opamp_top << c_route(pdk, cmirror_ibias.ports["A_gate_E"],cmirror_ibias.ports["B_gate_E"],extension=3*metal_sep,viaoffset=None)
    opamp_top << L_route(pdk, gate_short.ports["con_N"],cmirror_ibias.ports["A_drain_E"],viaoffset=False,fullbottom=False)
    srcshort = opamp_top << c_route(pdk, cmirror_ibias.ports["A_source_W"],cmirror_ibias.ports["B_source_W"],extension=metal_sep)
    opamp_top << straight_route(pdk, srcshort.ports["con_N"], cmirror_ibias.ports["welltie_N_top_met_S"],via2_alignment_layer="met2")
    # Route all tap rings together and ground them
    opamp_top << straight_route(pdk, cmirror_ibias.ports["welltie_N_top_met_N"],amp_fet_ref.ports["tie_S_top_met_S"])
    opamp_top << L_route(pdk, cmirror_ibias.ports["welltie_S_top_met_S"], opamp_top.ports["pin_gnd_E"],hwidth=4)
    # add ports, add bias/output pin, and return
    psuedo_out_port = movex(amp_fet_ref.ports["multiplier_0_source_E"].copy(),6*metal_sep)
    output_pin = opamp_top << straight_route(pdk, amp_fet_ref.ports["multiplier_0_source_E"], psuedo_out_port)
    opamp_top.add_ports(amp_fet_ref.get_ports_list(),prefix="outputstage_amp_")
    opamp_top.add_ports(cmirror_ibias.get_ports_list(),prefix="outputstage_bias_")
    opamp_top.add_ports(output_pin.get_ports_list(),prefix="pin_output_")
    bias_pin = opamp_top << rectangle(size=(5,3),layer=pdk.get_glayer("met3"),centered=True)
    bias_pin.movex(cmirror_ibias.center[0]).movey(cmirror_ibias.ports["B_gate_S"].center[1]-bias_pin.ymax-5*metal_sep)
    opamp_top << straight_route(pdk, bias_pin.ports["e2"], cmirror_ibias.ports["B_gate_S"],width=1)
    opamp_top.add_ports(bias_pin.get_ports_list(),prefix="pin_outputibias_")

    output_stage_netlist = opamp_output_stage_netlist(pdk, amp_fet_ref, biasParams)
    return opamp_top, output_stage_netlist

def opamp_netlist(two_stage_netlist: Netlist, output_stage_netlist: Netlist) -> Netlist:
    top_level_netlist = Netlist(
        circuit_name="opamp",
        nodes=["CSoutput", "vdd", "plus", "minus", "commonsourceibias", "outputibias", "diffpairibias", "gnd", "output"]
    )

    top_level_netlist.connect_netlist(
        two_stage_netlist,
        [('VDD', 'vdd'), ('GND', 'gnd'), ('DIFFPAIR_BIAS', 'diffpairibias'), ('VP', 'plus'), ('VN', 'minus'), ('CS_BIAS', 'commonsourceibias'), ('VOUT', 'CSoutput')]
    )

    top_level_netlist.connect_netlist(
        output_stage_netlist,
        [('VDD', 'vdd'), ('GND', 'gnd'), ('IBIAS', 'outputibias'), ('VIN', 'CSoutput'), ('VOUT', 'output')]
    )

    return top_level_netlist

@cell
def opamp(
    pdk: MappedPDK,
    half_diffpair_params: tuple[float, float, int] = (6, 1, 4),
    diffpair_bias: tuple[float, float, int] = (6, 2, 4),
    half_common_source_params: tuple[float, float, int, int] = (7, 1, 10, 3),
    half_common_source_bias: tuple[float, float, int, int] = (6, 2, 8, 2),
    output_stage_params: tuple[float, float, int] = (5, 1, 16),
    output_stage_bias: tuple[float, float, int] = (6, 2, 4),
    half_pload: tuple[float,float,int] = (6,1,6),
    mim_cap_size=(12, 12),
    mim_cap_rows=3,
    rmult: int = 2,
    with_antenna_diode_on_diffinputs: int=5
) -> Component:
    """
    create a two stage opamp with an output buffer, args->
    pdk: pdk to use
    half_diffpair_params: diffpair (width,length,fingers)
    diffpair_bias: bias transistor for diffpair nmos (width,length,fingers). The ref and output of the cmirror are identical
    half_common_source_params: pmos top component amp (width,length,fingers,mults)
    half_common_source_bias: bottom L/R large nmos current mirror (width,length,fingers,mults). The ref of the cmirror always has 1 multplier. multiplier must be >=2
    ****NOTE: change the multiplier option to change the relative sizing of the current mirror ref/output
    output_stage_amp_params: output amplifier transistor params (width, length, fingers)
    output_stage_bias: output amplifier current mirror params (width, length, fingers). The ref and output of the cmirror are identical
    half_pload: all 4 pmos load transistors of first stage (width,length,...). The last element in the tuple is the fingers of the bottom two pmos.
    mim_cap_size: width,length of individual mim_cap
    mim_cap_rows: number of rows in the mimcap array (always 2 cols)
    rmult: routing multiplier (larger = wider routes)
    with_antenna_diode_on_diffinputs: adds antenna diodes with_antenna_diode_on_diffinputs*(1um/0.5um) on the positive and negative inputs to the opamp
    """
    opamp_top = opamp_twostage(
        pdk,
        half_diffpair_params,
        diffpair_bias,
        half_common_source_params,
        half_common_source_bias,
        half_pload,
        mim_cap_size,
        mim_cap_rows,
        rmult,
        with_antenna_diode_on_diffinputs
    )
    # add output amplfier stage
    opamp_top, output_stage_netlist = __add_output_stage(pdk, opamp_top, output_stage_params, output_stage_bias, rmult)
    opamp_top.info['netlist'] = opamp_netlist(opamp_top.info['netlist'], output_stage_netlist)

    # return
    return rename_ports_by_orientation(component_snap_to_grid(opamp_top))





from gdsfactory.typings import Component
from pydantic import validate_arguments


@validate_arguments
def component_snap_to_grid(comp: Component) -> Component:
	"""snaps all polygons and ports in component to grid
	comp = the component to snap to grid
	NOTE this function will flatten the component
	"""
	#return comp.flatten()
	# flatten the component then copy (snaps polygons and ports to grid)
	name = comp.name
	comp = comp.flatten().copy()
	comp.name = name
	return comp





from glayout.pdk.mappedpdk import MappedPDK
from pydantic import validate_arguments
from gdsfactory.component import Component
from typing import Callable
from glayout.primitives.fet import nmos, pmos
from glayout.pdk.util.comp_utils import evaluate_bbox

@validate_arguments
def two_transistor_place(pdk: MappedPDK, pattern: str, deviceA: tuple[Callable, dict], deviceB: tuple[Callable, dict]) -> Component:
    """Place two transitors according to the patter provided
    args:
    pdk = MappedPDK to use
    pattern = placement pattern. This string must contain only white space, the char a, the char b.
    **** any other chars result in error. White space indicates a new row in the place
    **** all rows must have same number of cols
    deviceA/deviceB = tuple(function to call, kwargs for function) kwargs must include pdk
    """
    toplvlcomp = Component("2tranplace")
    # create the transistors
    tranA = deviceA[0](**deviceA[1])
    tranA_dims = evaluate_bbox(tranA)
    tranB = deviceB[0](**deviceB[1])
    tranB_dims = evaluate_bbox(tranB)
    # parse pattern into a matrix
    pattern = pattern.lower().split()
    parsed_pattern = list()
    for i, row in enumerate(pattern):
        parsed_pattern.append(list())
        if i==0:
            num_cols = len(row)
        elif len(row)!=num_cols:
            raise ValueError("all rows should have same number of devices")
        for char in row:
            if char=="a":
                parsed_pattern[i].append(tranA)
            elif char=="b":
                parsed_pattern[i].append(tranB)
            else:
                raise ValueError("pattern should only contain a,b, or whitespace")
    # run place (center, then right, then left, ...)
    extra_sep = 2*pdk.util_max_metal_seperation()
    yspace = extra_sep + max(tranA_dims[1], tranB_dims[1])
    xspace = extra_sep + max(tranA_dims[0], tranB_dims[0])
    for i, row in enumerate(parsed_pattern):
        for j, tran in enumerate(row):
            tranref = toplvlcomp << tran
            ymov = i * yspace
            xmov = j * xspace * (-1**(j%2))
            tranref.movex(xmov).movey(ymov)
            toplvlcomp.add_ports(tranref.get_ports_list(), prefix=f"place{i}_{j}_")
    return toplvlcomp



"""
usage: from gf180_mapped import gf180_mapped_pdk
"""

from gf180.layers import LAYER  # , LAYER_VIEWS
from ..gf180_mapped.grules import grulesobj
from ..mappedpdk import MappedPDK
from pathlib import Path

LAYER = LAYER.dict()
#LAYER["fusetop"]=(75, 0)
LAYER["CAP_MK"] = (117,5)

gf180_glayer_mapping = {
    "met5": "metal5",
    "via4": "via4",
    "met4": "metal4",
    "via3": "via3",
    "met3": "metal3",
    "via2": "via2",
    "met2": "metal2",
    "via1": "via1",
    "met1": "metal1",
    "mcon": "contact",
    "poly": "poly2",
    "active_diff": "comp",
    "active_tap": "comp",
    "n+s/d": "nplus",
    "p+s/d": "pplus",
    "nwell": "nwell",
    "pwell": "lvpwell",
    "dnwell": "dnwell",
    "capmet": "CAP_MK"
}

# note for DRC, there is mim_option 'A'. This is the one configured for use

gf180_lydrc_file_path = Path(__file__).resolve().parent / "gf180mcu_drc.lydrc"


gf180_mapped_pdk = MappedPDK(
    name="gf180",
    glayers=gf180_glayer_mapping,
	models={
        'nfet': 'nfet_03v3',
		'pfet': 'pfet_03v3',
		'mimcap': 'mimcap_1p0fF'
    },
    layers=LAYER,
    #klayout_lydrc_file=gf180_lydrc_file_path,
    grules=grulesobj,
)

# configure the grid size and other settings
gf180_mapped_pdk.gds_write_settings.precision = 5*10**-9
gf180_mapped_pdk.cell_decorator_settings.cache=False



from pydantic import validate_arguments
from gdsfactory.snap import snap_to_grid
from gdsfactory.typings import Component, ComponentReference
from gdsfactory.components.rectangle import rectangle
from gdsfactory.port import Port
from typing import Callable, Union, Optional, Iterable
from decimal import Decimal
from gdsfactory.functions import transformed
from gdsfactory.functions import move as __gf_move
from glayout.pdk.mappedpdk import MappedPDK
from gdstk import rectangle as primitive_rectangle
from .port_utils import add_ports_perimeter, rename_ports_by_list


@validate_arguments
def evaluate_bbox(custom_comp: Union[Component, ComponentReference], return_decimal: Optional[bool]=False, padding: float=0) -> tuple[Union[float,Decimal],Union[float,Decimal]]:
	"""returns the length and height of a component like object"""
	compbbox = custom_comp.bbox
	width = abs(Decimal(str(compbbox[1][0])) - Decimal(str(compbbox[0][0]))) + 2*Decimal(str(padding))
	height = abs(Decimal(str(compbbox[1][1])) - Decimal(str(compbbox[0][1]))) + 2*Decimal(str(padding))
	if return_decimal:
		return (width,height)
	return (float(width),float(height))


@validate_arguments
def move(custom_comp: Union[Port, ComponentReference, Component], offsetxy: tuple[float,float] = (0,0), destination: Optional[tuple[Optional[float],Optional[float]]]=None, layer: Optional[tuple[int,int]]=None) -> Union[Port, ComponentReference, Component]:
	"""moves custom_comp
	moves by offset[0]=x offset, offset[1]=y offset
	destination (x,y) if not none overrides offset option
	layer if specfied will move based on a layer (only relevant for destination option)
	returns the modified custom_comp
	"""
	if layer and isinstance(custom_comp, Component):
		custom_comp_ext = custom_comp.extract(layers=[layer,])
	elif layer and isinstance(custom_comp, ComponentReference):
		raise NotImplementedError("layer not implemented for comp ref")
	elif layer and isinstance(custom_comp,Port):
		raise TypeError("move:layer option for Port does not exist")
	else:
		custom_comp_ext = custom_comp
	if destination is not None:
		xoffset = destination[0] - custom_comp_ext.center[0] if destination[0] is not None else 0
		yoffset = destination[1] - custom_comp_ext.center[1] if destination[1] is not None else 0
	if isinstance(custom_comp, Port):
		if destination is None:
			custom_comp = custom_comp.move_copy(offsetxy)
		else:
			custom_comp = custom_comp.move_copy((xoffset,yoffset))
	elif isinstance(custom_comp, ComponentReference):
		if destination is None:
			custom_comp.movex(offsetxy[0]).movey(offsetxy[1])
		else:
			custom_comp.movex(xoffset).movey(yoffset)
	elif isinstance(custom_comp, Component):
		ref = custom_comp.copy().ref()
		if destination is None:
			ref.movex(offsetxy[0]).movey(offsetxy[1])
		else:
			ref.movex(xoffset).movey(yoffset)
		custom_comp = transformed(ref).copy()
	return custom_comp


@validate_arguments
def movex(custom_comp: Union[Port, ComponentReference, Component], offsetx: Optional[float] = 0, destination: Optional[float]=None, layer: Optional[tuple[int,int]]=None) -> Union[Port, ComponentReference, Component]:
	"""moves custom_comp by offsetx in the x direction
	returns the modified custom_comp
	"""
	if destination is not None:
		destination = (destination, None)
	return move(custom_comp, (offsetx,0),destination,layer)


@validate_arguments
def movey(custom_comp: Union[Port, ComponentReference, Component], offsety: Optional[float] = 0, destination: Optional[float]=None, layer: Optional[tuple[int,int]]=None) -> Union[Port, ComponentReference, Component]:
	"""moves custom_comp by offsety in the y direction
	returns the modified custom_comp
	"""
	if destination is not None:
		destination = (None, destination)
	return move(custom_comp, (0,offsety),destination,layer)


@validate_arguments
def align_comp_to_port(
	custom_comp: Union[Component,ComponentReference],
	align_to: Port,
	alignment: Optional[tuple[Optional[str],Optional[str]]] = None,
	layer: Optional[tuple[int,int]] = None,
	rtr_comp_ref = True
) -> Union[Component,ComponentReference]:
	"""Returns component/componentReference of component/componentReference aligned to port as specifed
	by default returns a componentReference
	for componentReference, the componentReference is modified (mutable), but for component, a copy of component is returned
	args:
	custom_comp = component to align properly
	align_to = Port to align to
	alignment = tuple(str,str) = (xalign,yalign). You can individually specify x/y algin=None and that means do nothing for that dim
	***NOTE, if left None, function will align component to outside and center of port (based on port orientation), specify (None,None) for real no align (do not move at all)
	****xalign = either l/left or r/right or c/center or None. component will be flush to right or left side of port or centered
	****yalgin = either t/top or b/bottom or c/center or None. top or bottom edge or center of component will align with port top/bottom/center
	layer = extract this layer from the component and aligns to this layer.
	rtr_comp_ref = will return a component reference if set true, else return component
	"""
	# find center and bbox
	if isinstance(custom_comp, ComponentReference):
		comp_type = transformed(custom_comp)
	else:
		comp_type = custom_comp
	if layer:
		comp_type = comp_type.extract([layer])
	cbbox = comp_type.bbox
	ccenter = comp_type.center
	# setup
	xdim = abs(cbbox[1][0] - cbbox[0][0])
	ydim = abs(cbbox[1][1] - cbbox[0][1])
	width = align_to.width
	is_EW = bool(round(align_to.orientation + 90) % 180) # not EW == NS
	# error checks and decide orientation if None
	if alignment is None:
		if round(align_to.orientation) == 0:# facing east
			xalign = "r"
			yalign = "c"
		elif round(align_to.orientation) == 180:# facing west
			xalign = "l"
			yalign = "c"
		elif round(align_to.orientation) == 270:# facing south
			xalign = "c"
			yalign = "b"
		elif round(align_to.orientation) == 90:#facing north
			xalign = "c"
			yalign = "t"
		else:
			raise ValueError("port must be vertical or horizontal")
	else:
		xalign = (alignment[0] or "none").lower()
		yalign = (alignment[1] or "none").lower()
	# compute translation x amount for x alignment
	x_movcenter = align_to.center[0] - ccenter[0]
	x_mov_lr = abs(xdim/2 if is_EW else (width-xdim)/2)
	if "none" in xalign:
		xmov = 0
	elif "l" in xalign:
		xmov = x_movcenter - x_mov_lr
	elif "r" in xalign:
		xmov = x_movcenter + x_mov_lr
	elif "c" in xalign:
		xmov = x_movcenter
	else:
		raise ValueError("please specify valid x alignment of l/r/c/None")
	# compute translation y amount for y alignment
	y_movcenter = align_to.center[1] - ccenter[1]
	y_move_updown = abs((width-ydim)/2 if is_EW else ydim/2)
	if "none" in yalign:
		ymov = 0
	elif "t" in yalign:
		ymov = y_movcenter + y_move_updown
	elif "b" in yalign:
		ymov = y_movcenter - y_move_updown
	elif "c" in yalign:
		ymov = y_movcenter
	else:
		raise ValueError("please specify valid y alignment of t/b/c/None")
	# make reference type, execute move
	if isinstance(custom_comp, Component):
		comp_ref = custom_comp.ref()
	else:
		comp_ref = custom_comp
	comp_ref.movex(xmov).movey(ymov)
	# make correct type and return
	if rtr_comp_ref:
		return comp_ref
	else:
		return transformed(comp_ref)


@validate_arguments
def to_decimal(elements: Union[tuple,list,float,int,str]):
	"""converts all elements of list like object into decimals
	or converts single num into decimal"""
	if not isinstance(elements,Iterable):
		return Decimal(str(elements))
	else:
		elements = list(elements)
	for i, element in enumerate(elements):
		if isinstance(element,Union[int,float]):
			elements[i] = Decimal(str(element))
	return elements

@validate_arguments
def to_float(elements: Union[tuple,list,Decimal,float]):
	"""converts all elements of list like object into floats and snaps to grid
	or converts single decimal into floats"""
	if not isinstance(elements,Iterable):
		return snap_to_grid(float(elements))
	else:
		elements = list(elements)
	for i, element in enumerate(elements):
		if isinstance(element, Union[float,Decimal]):
			elements[i] = snap_to_grid(float(element))
	return elements

@validate_arguments
def prec_array(custom_comp: Component, rows: int, columns: int, spacing: tuple[Union[float,Decimal],Union[float,Decimal]], absolute_spacing: Optional[bool]=False) -> Component:
	"""instead of using the component.add_array function, if you are having grid snapping issues try using this function
	works the same way as add_array but uses decimals and snaps to grid to mitigate grid snapping issues
	args
	custom_comp: Component type to make an array from
	columns: num cols in the array
	rows: num rows in the array
	absolute_spacing: the spacing mode of spacing variable
	spacing: IF absolute_spacing spacing BETWEEN elements in the array ELSE spacing BETWEEN ORIGINS of elements in the array
	****NOTE do not use negative spacing, instead specify absolute_spacing=True
	"""
	# make sure to work with decimals
	precspacing = list(spacing)
	for i in range(2):
		if isinstance(spacing[i],Union[int,float]):
			precspacing[i] = Decimal(str(spacing[i]))
	if not absolute_spacing:
		precspacing = [precspacing[i] + evaluate_bbox(custom_comp,True)[i] for i in range(2)]
	# create array
	precarray = Component()
	for colnum in range(columns):
		coldisp = colnum * precspacing[0]
		for rownum in range(rows):
			rowdisp = rownum * precspacing[1]
			cref = precarray << custom_comp
			cref.movex(to_float(coldisp)).movey(to_float(rowdisp))
			precarray.add_ports(cref.get_ports_list(),prefix=f"row{rownum}_col{colnum}_")
	return precarray.flatten()


@validate_arguments
def prec_center(custom_comp: Union[Component,ComponentReference], return_decimal: bool=False) -> tuple[Union[float,Decimal],Union[float,Decimal]]:
	"""instead of using component.ref_center() to get the center of a component,
	use this function which will return the correct offset to center a component
	returns (x,y) corrections
	if return_decimal=True, return in Decimal, otherwise return float"""
	correctmax = [dim/2 for dim in evaluate_bbox(custom_comp, True)]
	currentmax = to_decimal((custom_comp.xmax,custom_comp.ymax))
	correctionxy = [correctmax[i] - currentmax[i] for i in range(2)]
	if return_decimal:
		return correctionxy
	return to_float(correctionxy)

@validate_arguments
def prec_ref_center(custom_comp: Union[Component,ComponentReference], destination: Optional[tuple[float,float]]=None) -> ComponentReference:
	"""instead of using component.ref_center() to get a ref to center at origin,
	use this function which will return a centered ref
	you can then run component.add(prec_ref_center(custom_comp)) to add the reference to your component
	you can optionally specify a destination = tuple(x,y), by default, the destination is 0,0
	returns component reference
	"""
	compref = custom_comp if isinstance(custom_comp, ComponentReference) else custom_comp.ref()
	xcor, ycor = prec_center(compref, False)
	if destination is not None:
		xcor += destination[0]
		ycor += destination[1]
	return compref.movex(xcor).movey(ycor)



def get_padding_points_cc(
	custom_comp: Union[ComponentReference, Component, list],
	default: float = 50.0,
	top: Optional[float]=None,
	bottom: Optional[float]=None,
	right: Optional[float]=None,
	left: Optional[float]=None,
	pdk_for_snap2xgrid: Optional[MappedPDK]=None
) -> list:
	"""works like gdsfactory.add_padding.get_padding_points, but also accepts componentReference or bbox
	additionally, if you optionally pass a pdk it will snap to 2x grid (else just operates like get_padding_points)"""
	if isinstance(custom_comp, ComponentReference) or isinstance(custom_comp, Component):
		bbox = custom_comp.bbox
	else:
		bbox = custom_comp
	top = top if top is not None else default
	bottom = bottom if bottom is not None else default
	right = right if right is not None else default
	left = left if left is not None else default
	ppoints = [
		[bbox[0][0] - left, bbox[0][1] - bottom],
		[bbox[1][0] + right, bbox[0][1] - bottom],
		[bbox[1][0] + right, bbox[1][1] + top],
		[bbox[0][0] - left, bbox[1][1] + top],
	]
	if pdk_for_snap2xgrid is not None:
		for i, ppoint in enumerate(ppoints):
			ppoints[i] = pdk_for_snap2xgrid.snap_to_2xgrid(ppoint)
	return ppoints



def get_primitive_rectangle(size: tuple[float,float]=(5,3), layer: tuple[int,int]=(0,0)):
	"""creates a rectangle component which snaps point to grid (does not snap to 2x grid)
	has same behavoir as gdsfactory.components.rectangle but doesnt allow centering (would snap to grid)
	"""
	temprect = Component()
	temprect.add_polygon(primitive_rectangle((0,0),size,*layer))
	temprect = rename_ports_by_list(add_ports_perimeter(temprect,layer=layer,prefix="route_"),[("W","e1"),("N","e2"),("E","e3"),("S","e4")])
	#rect = Component()
	#clogic_ref = prec_ref_center(temprect) if centered else temprect.ref()
	#rect.add(clogic_ref)
	#rect.add_ports(clogic_ref.ports)
	return temprect.flatten()
